import re
import spacy
from naimai.constants.regex import regex_eft, regex_abbrvs,regex_abbrvs2, regex_some_brackets,\
    regex_numbers_in_brackets, regex_equation, regex_etal, regex_equation_tochange, regex_explained_2pts
from naimai.utils.regex import remove_between_brackets
from naimai.constants.nlp import nlp_vocab

print('nlp loaded for TextCleaner')
print('')
loaded_nlp = spacy.load(nlp_vocab)

class TextCleaner:
    def __init__(self, text,nlp=None):
        self.text = text
        self.cleaned_text = ''
        if nlp:
            self.nlp = nlp
        else:
            self.nlp = loaded_nlp

    def remove_explanations(self,text):
        '''
        remove text after ':' that explains stuff.
        :param text:
        :return:
        '''
        return re.sub(regex_explained_2pts,'.',text)
    def replace_abbrevs(self,text):
        '''
        replace i.e. and e.g. by meaning in the text & cf. by "see" & et al. by et al
        :param text:
        :return:
        '''
        text1= re.sub(regex_abbrvs,'meaning', text)
        text2 = re.sub(regex_etal,'et al', text1)
        return re.sub(regex_abbrvs2,'see', text2)

    def remove_some_brackets(self,text):
        '''
        remove brackets with > 5 words & brackets and square brackets with words figure, table, meaning (that would have replaced i.e. and
        e.g. using replace_abbrevs)
        :param text:
        :return:
        '''
        txt2 = remove_between_brackets(text)
        return re.sub(regex_some_brackets,'',txt2, flags=re.I)


    def remove_sentence_with_nbs_brackets(self,text):
        '''
        remove sentences that contains number in between brackets & square brackets
        :param text:
        :return:
        '''
        return re.sub(regex_numbers_in_brackets, '', text)

    def remove_sentence_with_eft(self,text):
        '''
        remove sentences that contains equations & the eft terms : equation, figure or table!
        :param text:
        :return:
        '''
        text_stacked_eqs = re.sub(regex_equation_tochange,r'\1=\2',text)
        doc = self.nlp(text_stacked_eqs)
        sentences = [sent.text.strip() for sent in doc.sents]
        sentences_filtered1 = [stc for stc in sentences if not re.findall(regex_equation, stc, flags=re.I)] # remove sentences with eqs
        sentences_filtered2 = [stc for stc in sentences_filtered1 if not re.findall(regex_eft,stc,flags=re.I)] # remove sentences with eft
        sentences_filtered3 = [stc for stc in sentences_filtered2 if len(stc.split())>3]
        cleaned_text = '. '.join(sentences_filtered3)
        return cleaned_text

    def fix_spaces(self,text):
        '''
        keep only one space between words & replace 2 points in row
        :param text:
        :return:
        '''
        txt2 = re.sub('\.\s*\.','.',text)
        return re.sub('\s+',' ', txt2)

    def clean(self):
        '''
        clean the text by :
            1. remove explained text that happens after ':'
            2. replacing i.e. & e.g. abbreviations by "meaning"
            3. removing brackets with table, figure
            4. removing sentences with [x] or (x), x being a digit
            5. removing sentences with equations, figure or table terms
            6. fixing additional spaces & other stuff resulted by cleaning methods
        :return:
        '''
        self.text = self.remove_explanations(self.text)
        self.text = self.replace_abbrevs(self.text)
        self.text = self.remove_some_brackets(self.text)
        self.text = self.remove_sentence_with_nbs_brackets(self.text)
        self.text = self.remove_sentence_with_eft(self.text)
        self.cleaned_text = self.fix_spaces(self.text)






import re
from io import StringIO, BytesIO
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.converter import TextConverter
from pdfminer.layout import LAParams
from pdfminer.pdfpage import PDFPage
import pdf2image
import pytesseract
import unidecode


def convert_pdf_to_txt(obj,is_path=True,use_ocr=False):
    result = ''
    try:
        result=pm_convert_pdf_to_txt(obj,is_path=is_path)
        nb_new_lines = len(re.findall('\n[^.?]\n', result))
        if (nb_new_lines >5000) or nb_new_lines==0:
            result = ''
            if is_path:
                print('OCR TO USE in ', obj)
            else:
                print('OCR USED')
            if use_ocr:
                result = ocr_convert_pdf_to_txt(obj,is_path=is_path)

    except:
        if is_path:
            print('except OCR USED in ', obj)
        else:
            print('OCR USED')
        if use_ocr:
            result = ocr_convert_pdf_to_txt(obj,is_path=is_path)

    return result

def ocr_convert_pdf_to_txt(obj,is_path=True):
    if is_path:
        with open(obj, 'rb') as f:
            pdf_content = f.read()
    else:
        pdf_content = BytesIO(obj)

    doc = pdf2image.convert_from_bytes(pdf_content)
    article = []
    for page_data in doc:
        txt = pytesseract.image_to_string(page_data).encode("utf-8")
        article.append(txt.decode("utf-8"))
    text = " ".join(article)
    return unidecode.unidecode(text.replace('e´', 'é'))

def pm_convert_pdf_to_txt(obj, pages=None,is_path=True):
    if not pages:
        pagenums = set()
    else:
        pagenums = set(pages)
    output = StringIO()
    manager = PDFResourceManager()
    converter = TextConverter(manager, output, laparams=LAParams())
    interpreter = PDFPageInterpreter(manager, converter)

    if is_path:
        infile = open(obj, 'rb')
    else:
        infile = BytesIO(obj)
    for page in PDFPage.get_pages(infile, pagenums):
        interpreter.process_page(page)
    infile.close()
    converter.close()
    text = output.getvalue()
    output.close()
    return unidecode.unidecode(text.replace('e´','é'))

def find_in_txt(original,lower_txt):
  idx1=original.lower().find(lower_txt)
  idxN = idx1 + len(lower_txt)
  return original[idx1:idxN]



from transformers import AutoModelForSequenceClassification, AutoTokenizer
from naimai.constants.paths import path_objective_classifier
import torch

class Objective_classifier:
    def __init__(self,model=None,tokenizer=None):
        if model and tokenizer:
            self.model= model
            self.tokenizer =  tokenizer
        else:
            self.model = AutoModelForSequenceClassification.from_pretrained(path_objective_classifier, num_labels=2)
            self.tokenizer = AutoTokenizer.from_pretrained(path_objective_classifier)
            if torch.cuda.is_available():
                print('  >> GPU Used in objective classification !')
                self.model = self.model.to('cuda')

    def sentence_is_objective(self,sentence):
        txt_tokenized = self.tokenizer(sentence, truncation=True, padding=True, return_tensors='pt')
        if torch.cuda.is_available():
            txt_tokenized=txt_tokenized.to("cuda")
        outputs = self.model(**txt_tokenized)
        probs = outputs[0].softmax(1)
        idx = probs.argmax()
        prediction = int(idx.cpu().numpy())
        if prediction > 0:
            return True
        return False

    def predict(self,list_objectives):
        result = [sentence for sentence in list_objectives if self.sentence_is_objective(sentence)]
        return result


''' 
- paper_base Class: used extract data about a paper and format it as a dictionary. Each paper data is contained in a
 row of a csv file. The data are: :  doi, fields, abstract, title, authors, year of publication, journal, num of papers
 that cited the papers (numCitedBy).

- papers Class: used to format the input csv file that contains data about many papers. Each paper is formatted using
  the paper_bass class. In case a meta data needs to be overwritten, a new 'paper' and 'papers' class can be implemented, that inherits from
  paper_base and papers, as in the files in papers/only_abstracts/.

- paper_full_base Class [Not Finished]: class used to extract data from all
the article (instead of only the abstract, as in paper_base). 
'''

import random
import os
from naimai.utils.general import load_gzip, save_gzip
from naimai.constants.paths import naimai_dois_path
from naimai.models.abbreviation import extract_abbreviation_definition_pairs
# from naimai.processing import TextCleaner
from tqdm.notebook import tqdm
import pandas as pd
import numpy as np
from ast import literal_eval
import spacy
from naimai.constants.paths import path_open_citations
from naimai.utils.general import get_soup
from naimai.constants.nlp import nlp_vocab

from naimai.utils.regex import multiple_replace
from naimai.decorators import update_naimai_dois

from spacy.language import Language
from spacy_langdetect import LanguageDetector


def create_lang_detector(nlp, name):
    return LanguageDetector()

class paper_base:
    '''
    Paper class that map a row if the csv file with data about papers to a dictionary.
    '''
    def __init__(self,df: pd.DataFrame,idx_in_df: int):
        '''
        :param df: dataframe of the csv file with many papers data
        :param idx_in_df: index of row in df containing a paper data
        '''
        self.database=''
        self.file_name = idx_in_df
        self.paper_infos = df.iloc[idx_in_df,:]
        self.raw_text = ''
        self.fields = []
        self.numCitedBy = .5
        self.Introduction = ''
        self.Journal = ''
        self.highlights = []
        self.Abstract = ''
        self.Conclusion = ''
        self.Keywords = ''
        self.Authors = ''
        self.Title = ''
        self.year = 999
        self.References = []
        self.doi = ''


    def get_abbreviations_dict(self) -> dict:
        '''
        Get abbreviations and their meanings from the title and the abstract in a dictionary.
        Modified code from : https://gist.github.com/ijmarshall/b3d1de6ccf4fb8b5ee53
        :return:
        '''
        abstract_abbrevs = extract_abbreviation_definition_pairs(doc_text=self.Abstract)
        intro_abbrevs = extract_abbreviation_definition_pairs(doc_text=self.Introduction)
        abstract_abbrevs.update(intro_abbrevs)

        corrected_abbrevs = {}
        for k in abstract_abbrevs:
            corrected_abbrevs[' ' + k] = ' ' + abstract_abbrevs[k] + ' ' + '(' + k + ')'
        return corrected_abbrevs

    def replace_abbreviations(self):
        '''
        Replace each abbreviation (in title and abstract) by its meaning.
        :return:
        '''
        abbreviations_dict = self.get_abbreviations_dict()
        if abbreviations_dict:
            self.Abstract = multiple_replace(abbreviations_dict, self.Abstract)
            self.Title = multiple_replace(abbreviations_dict, self.Title)

    # def clean_text(self,text):
    #     '''
    #     clean the text based on TextCleaner object
    #     :param text:
    #     :return:
    #     '''
    #     cleaner = TextCleaner(text)
    #     cleaner.clean()
    #     return cleaner.cleaned_text

    def is_in_database(self,list_dois: list) -> bool:
        '''
        Check if the paper is already processed (its doi is in naimai database)
        :param list_dois: list of naimai dois
        :return:
        '''
        if self.doi in list_dois:
            return True
        return False

    def is_paper_english(self,nlp) -> bool:
        '''
        Detect language if paper language is english based on its title.
        :param nlp: spaCy nlp pipeline
        :return:
        '''
        if self.Title:
            title_nlp = nlp(self.Title)
            language_score=title_nlp._.language
            condition_english = (language_score['language']=='en') and (language_score['score']>0.7)
            if condition_english:
                return True
            else:
                return False
        return True

    def get_doi(self):
        '''
        Get doi of the paper
        :return:
        '''
        self.doi = self.paper_infos['doi']

    def get_fields(self) :
        '''
        Get list fields of the paper
        :return:
        '''
        self.fields = literal_eval(self.paper_infos['fields'])

    def get_Abstract(self):
        '''
        Get the abstract of the paper
        :return:
        '''
        abstract = self.paper_infos['abstract']
        if isinstance(abstract,str):
            self.Abstract = abstract.replace('-\n', '').replace('\n', ' ')
        else:
            self.Abstract = ''

    def get_Title(self):
        '''
        Get the title of the paper
        :return:
        '''
        self.Title = self.paper_infos['title'].replace('-\n', '').replace('\n', ' ')

    def get_Authors(self):
        '''
        Get the authors of the paper (in str format)
        :return:
        '''
        self.Authors = self.paper_infos['authors'].replace(';',',')

    
    def get_year(self):
        '''
        Get the year of publication of the paper
        :return:
        '''
        try:
            self.year = int(self.paper_infos['date'])
        except:
            self.year = ''
            print('Year to be corrected in ', self.doi)

    def save_dict(self) -> dict:
        '''
        Format the read data of the paper into a dictionary
        :return:
        '''
        attr_to_save = ['doi', 'Authors', 'year','database','fields','Abstract','Keywords', 'Title','numCitedBy','Journal']
        paper_to_save = {key: self.__dict__[key] for key in attr_to_save}
        return paper_to_save

    def get_journal(self):
        '''
        Get the journal of the paper
        :return:
        '''
        self.Journal = self.paper_infos['journals']

    def get_numCitedBy(self):
        '''
        Get the num of articles that cited this paper.
        :return:
        '''
        if 'numCitedBy' in self.paper_infos:
            self.numCitedBy = self.paper_infos['numCitedBy']
            if np.isnan(self.numCitedBy):
                self.numCitedBy = 0.5    
        else:
            path = path_open_citations + self.doi
            soup = get_soup(path)
            soup_list = literal_eval(soup.text)
            if isinstance(soup_list, list):
                self.numCitedBy = len(soup_list)

class papers:
    '''
    Class to process a csv file containing many papers data.
    '''
    def __init__(self, papers_path: str,database: str,nlp=None):
        '''
        :param papers_path: path of the csv file
        :param database: name of the database if all the papers in the csv file are coming from the same source.
        :param nlp:  spaCy nlp pipeline
        '''
        self.elements = {}
        self.database=database
        self.data = pd.read_csv(papers_path)
        print('Len data : ', len(self.data))

        # Getting nlp
        if nlp:
            self.nlp = nlp
        else:
            print('Loading nlp vocab..')
            self.nlp = spacy.load(nlp_vocab)
            Language.factory("language_detector", func=create_lang_detector)
            self.nlp.add_pipe('language_detector', last=True)

        # Getting naimai dois
        if os.path.exists(naimai_dois_path):
            self.naimai_dois = load_gzip(naimai_dois_path)
        else:
            print('No naimai dois..')
            self.naimai_dois=[]

    def __len__(self):
        return len(self.elements.keys())

    def __setitem__(self, key, value):
        self.elements[key] = value

    def __getitem__(self, item):
        return self.elements[item]

    def random_papers(self,k=3, seed=None) -> list:
        '''
        Get random papers from the csv file.
        :param k: num of papers wanted
        :param seed: seed of random
        :return:
        '''
        elts = list(self.elements)
        random.seed(seed)
        rds = random.sample(elts, k)
        papers_list = [self.elements[el] for el in rds]
        return papers_list

    def add_paper(self,idx_in_data: int,check_database=True):
        '''
        Add a paper data as element
        :param idx_in_data: idx of the paper in the csv file
        :param check_database: if True, the paper is not added as element if already contained in the database.
        :return:
        '''
        new_paper = paper_base(df=self.data,
                                idx_in_df=idx_in_data)
        new_paper.get_doi()
        new_paper.get_Title()
        if check_database:
            is_in_database_condition = new_paper.is_in_database(self.naimai_dois)
        else:
            is_in_database_condition = False
        if not is_in_database_condition:
            if new_paper.is_paper_english(self.nlp):
                new_paper.get_Abstract()
                if len(new_paper.Abstract.split()) > 5:
                    new_paper.database = self.database
                    new_paper.get_fields()
                    new_paper.get_journal()
                    new_paper.get_Authors()
                    new_paper.get_year()
                    new_paper.replace_abbreviations()
                    new_paper.get_numCitedBy()
                    self.elements[new_paper.doi] = new_paper.save_dict()
                    self.naimai_dois.append(new_paper.doi)

    @update_naimai_dois
    def get_papers(self,update_dois=False,idx_start=0,idx_finish=-1,show_tqdm=False,check_database=True):
        '''
        Get the papers from the csv file and format them as a dictionary, stored as 'element' in the instance.
        :param update_dois: if True, it's going to update the list of naimai dois when new papers are added.
        :param idx_start: if >0, the processing starts from the idx_start^th row
        :param idx_finish: if != -1, the processing finishes at the idx_finish^th row
        :param show_tqdm: if True, a progress bar is showed to track the processing
        :param check_database: if True, the paper is not added as element if already contained in the database.
        :return:
        '''
        if show_tqdm:
            range_ = tqdm(self.data.iterrows(),total=len(self.data))
        else:
            range_= self.data.iterrows()
        for idx,_ in range_:
            self.add_paper(idx_in_data=idx,check_database=check_database)
            

    def save_elements(self, file_dir: str,update=False):
        '''
        Saves the dictionary of processed documents
        :param file_dir: path where the processed elements are stored.
        :param update: if True, update the file_dir in case it already exists.
        :return:
        '''
        papers_to_save = self.__dict__['elements']
        if update and os.path.exists(file_dir):
            loaded_papers = load_gzip(file_dir)
            loaded_papers.update(papers_to_save)
            save_gzip(file_dir,loaded_papers)
        else:
            save_gzip(file_dir,papers_to_save)


    def update_naimai_dois(self):
        '''
        Update the naimai dois if new dois are processed
        :return:
        '''
        if self.naimai_dois:
            save_gzip(naimai_dois_path,self.naimai_dois)

class paper_full_base(paper_base):
    '''
    Class to process the full article (instead of the abstract). [Not Finished].
    '''
    def __init__(self):
        super().__init__()
        self.Introduction = {}
        self.Methods = {}
        self.Results = {}
        self.unclassified_section = {}

    def get_abbreviations_dict(self):
        '''
        get all abbreviations of the papers in Introduction & Methods sections.
        :return:
        '''
        abstract_abbrevs={}
        if isinstance(self.Abstract,str):
            abstract_abbrevs = extract_abbreviation_definition_pairs(doc_text=self.Abstract)
        elif isinstance(self.Abstract,dict):
            for elt in self.Abstract:
                abstract_abbrevs = extract_abbreviation_definition_pairs(doc_text=self.Abstract[elt])
                abstract_abbrevs.update(abstract_abbrevs)

        for elt in self.Introduction:
            intro_abbrevs = extract_abbreviation_definition_pairs(doc_text=self.Introduction[elt])
            abstract_abbrevs.update(intro_abbrevs)
        for elt in self.Methods:
            methods_abbrevs = extract_abbreviation_definition_pairs(doc_text=self.Methods[elt])
            abstract_abbrevs.update(methods_abbrevs)

        corrected_abbrevs = {}
        for k in abstract_abbrevs:
            corrected_abbrevs[' ' + k] = ' ' + abstract_abbrevs[k] + ' ' + '(' + k + ')'
        return corrected_abbrevs

    def is_paper_english(self,nlp) -> bool:
        '''
        Detect language if paper language is english based on its title.
        :return:
        '''
        if self.Title:
            title_nlp = nlp(self.Title)
            language_score=title_nlp._.language
            condition_english = (language_score['language']=='en') and (language_score['score']>0.7)
            if condition_english:
                return True
        return False

    def save_dict(self):
        attr_to_save = ['doi', 'Authors', 'year','database','fields','Abstract','Keywords','Title','numCitedBy', 'Journal']
        paper_to_save = {key: self.__dict__[key] for key in attr_to_save}
        return paper_to_save


replacement_words = {
        "-\n": '',
        "\n-": '',
        "\n\n": " ",
        "\n": " ",
        "1-D": "oneD",
        "1D": "oneD",
        "2-D": "twoD",
        "2D": "twoD",
        "3-D": "threeD",
        "3D": "threeD",
        "non uniform": "non-uniform",
        'process': '',
        'condition': '',
        'study': '',
        'method': '',
        'paper': '',
        'figure': '',
        'author': '',
        'CO2': 'COtwo',
        }

parsing_corrections={
        ' uid': ' fluid',
        ' uids': ' fluids',
        ' ow' : ' flow',
        ' ows': ' flows',
        ' ood': ' flood',
        ' oods': ' floods',
        ' ield': ' field',
        ' eld': ' field',
        ' ields': ' fields',
        ' inite': ' finite',
        'inuence' : 'influence',
        ' oodplain' : 'floodplain',
        'outow': 'outflow',
        ' irst': 'first',
}

reported_corrections = {
            ' - ': '-',
            ' – ': '–',
            ' , ': ', ',
            '( ': '(',
            ' )': ')',
            " ' ": "'",
            " / ": "/",
            ' these ': " the ",
            ' our ': ' their ',
            "the present ": " their ",
            "and and": 'and',
            'nonuniform': 'non-uniform',
            'also' : '',
            "In this ": 'in their ',
            ' this ': " their ",
            'in addition': '',
            '.we ': ".They ",
            ',we ': ",they ",
            ' we ': " they ",
            " us ": " them ",
            'Fig.': "",
            "Figure": "",
            '(i':'',
            '(e':"",
            ' above': '',
            ' cited ': ' ',
        }


import dask.bag as db
import json
import subprocess
# from tqdm.notebook import tqdm
from bs4 import BeautifulSoup
import re
from naimai.utils.regex import str_in_list
import os
# import tarfile

class ArXiv_Crawler:
  def __init__(self, arxiv_metadata_dir, tar_dir='',pdfs_dir='', pdf_manifest_dir='',category='',get_manifest=True):
    self.arxiv_metadata_dir = arxiv_metadata_dir
    self.pdf_manifest_dir = pdf_manifest_dir
    self.metadata_dic = {}
    self.all_ids = []
    self.ids_filtered = [] #ids without alphabets
    self.ids_not_found = []
    self.titles=[]
    self.pdfs_dir = pdfs_dir
    self.tar_dir = tar_dir
    self.category = category
    self.no_lettrs_pattern= '^[A-Za-z]'
    self.docs = None

    if get_manifest:
      self.get_metadata_dic_from_manifest()
      self.get_category_ids(self.category, take_all_category_files=True)


  def read(self):
    return db.read_text(self.arxiv_metadata_dir).map(json.loads)

  def get_category_docs(self,take_all_category_files, category):
    all_docs=self.read()
    docs= all_docs.filter(lambda x: x['categories']==category)
    filtered_docs= docs.filter(lambda x: x['comments']!='This paper has been withdrawn')
    if take_all_category_files:
      self.docs = filtered_docs
    else:
      pdf_files = os.listdir(self.pdfs_dir)
      self.docs = filtered_docs.filter(lambda x: x if str_in_list((x['id'] + '.pdf'),
                        pdf_files) else None)
      # self.docs = filtered_docs.filter(lambda x: x if (not str_in_list((x['title']+'_'+x['id']+'.pdf').replace('\n','_').replace('\\','').replace('/',''), pdf_files)) else None)

  def docs2df(self):
    return self.docs.to_dataframe()[['id','authors_parsed','title','abstract','doi']]

  def get_category_ids(self, category, take_all_category_files=True):
    self.get_category_docs(take_all_category_files, category)
    df_category = self.docs2df()
    self.all_ids = list(df_category['id'].apply(lambda x: str(x)))
    # title_list = list(df_category['title'])
    self.ids_filtered = [id for id in self.all_ids if not re.findall(self.no_lettrs_pattern, id)]
    print('  >> IDs = {}  -- IDs filtered = {}'.format(len(self.all_ids), len(self.ids_filtered)))
    # return {'ids': self.ids_filtered, 'titles': title_list}

  def get_metadata_dic_from_manifest(self):
    print('>> Constructing metadata dic')
    with open(self.pdf_manifest_dir, 'r') as manifest:
      soup = BeautifulSoup(manifest, 'xml')
      files = soup.find_all('file')

    for f in files:
      filepath = f.find('filename').text
      self.metadata_dic[filepath] = {'first_item': f.find('first_item').text,
                                'last_item': f.find('last_item').text,
                                'num_items': f.find('num_items').text,
                                'size': int(f.find('size').text) / 1000000000,
                                'yymm': f.find('yymm').text
                                }

  def find_id_tar_in_manifest(self, id_tofind): #id_tofind form : '0704.2668'
    # print('>> Find tar files')
    yymm_tofind = id_tofind.split('.')[0]
    number_tofind = int(id_tofind.split('.')[1])
    potential_paths = [path for path in self.metadata_dic if self.metadata_dic[path]['yymm'] == yymm_tofind]

    potential_paths_filtered = []
    for path in potential_paths:
      dic = self.metadata_dic[path]
      letter_in_items = re.findall(self.no_lettrs_pattern, dic['first_item']) + re.findall(self.no_lettrs_pattern, dic['last_item'])
      if not letter_in_items:
        yymm = dic['yymm']
        first_number = int(dic['first_item'].split('.')[1])
        last_number = int(dic['last_item'].split('.')[1])
        if number_tofind in range(first_number, last_number):
          tar_name = '_'.join([yymm, str(first_number), str(last_number)]) + '.tar'  # yymm_start_finish
          potential_paths_filtered.append({'path': path,
                                         'filename': tar_name})
    return potential_paths_filtered

  def download_tar(self, pdf_s3_path,output_path,output_filename):
    # print('>> tar download : ')
    output = os.path.join(output_path, output_filename)
    if output_filename not in os.listdir(output_path):
      cmd = 's3cmd get --recursive --requester-pays --skip-existing s3://arxiv/{} {}'.format(pdf_s3_path, output)
      subprocess.call(cmd,shell=True)

  def id_in_tar_dir(self,id):
    potential_paths_filtered = self.find_id_tar_in_manifest(id)
    if len(potential_paths_filtered) > 0:
      output_filename = potential_paths_filtered[0]['filename']
      if output_filename in os.listdir(self.tar_dir):
        return True
    return False

  def ids_in_tar_dir(self):
    result = {'ids found': [], 'ids not found': []}
    for id in self.ids_filtered:
      if self.id_in_tar_dir(id):
        result['ids found'].append(id)
      else:
        result['ids not found'].append(id)
    return result



  def download_tar_files(self):

    for id in self.ids_filtered:
      potential_paths_filtered = self.find_id_tar_in_manifest(id)
      if len(potential_paths_filtered)>0:
        pdf_s3_path = potential_paths_filtered[0]['path']
        output_path = self.tar_dir
        output_filename = potential_paths_filtered[0]['filename']
        self.download_tar(pdf_s3_path,output_path, output_filename)
        if len(potential_paths_filtered)>1:
          print('  >> LEN >1 FOR ID ', id)
      else:
        self.ids_not_found.append(id)
        print('  >> PROBLEM WITH ID ', id)


  # def download_lost_ids(self):
  #   useragent = UserAgent()
  #   for id in tqdm(self.ids_not_found):
  #     url = 'https://arxiv.org/pdf/'+str(id)
  #     ua = useragent.random
  #     response = requests.get(url, headers={'User-Agent': ua})
  #     with open(self.pdfs_dir, 'wb') as f:
  #       f.write(response.content)
  #     time.sleep(np.random.random() + 5)

  def tar2pdfs(self):
    for id in self.ids_filtered:
      self.get_id_file_from_tars(id)

  def get_id_file_from_tars(self,id):
    tar_files_dic = [{'yymm': elt.split('_')[0],
                      'first_item': elt.split('_')[1],
                      'last_item': elt.split('_')[2].replace('.tar', ''), }
                     for elt in os.listdir(self.tar_dir)]
    id_dic = {'yymm': id.split('.')[0], 'number': id.split('.')[1]}
    mytar=''
    for tar in tar_files_dic:
      if id_dic['yymm'] == tar['yymm']:
        if int(id_dic['number']) in range(int(tar['first_item']), int(tar['last_item'])):
          mytar = tar['yymm'] + '_' + tar['first_item'] + '_' + tar['last_item'] + '.tar'
          break

    # try:
    #   if mytar:
    #     tarfile_dir = os.path.join(self.tar_dir, mytar)
    #     tar = tarfile.open(tarfile_dir)
    #     tar_content = tar.getnames()
    #     id_in_tar = [elt for elt in tar_content if id in elt]
    #
    #     if id_in_tar:
    #       id_in_tar = id_in_tar[0]
    #       if id_in_tar not in self.pdfs_dir:
    #         tar.extract(id_in_tar, self.pdfs_dir)
    # except:
    #   print('problem in id : {}, mytar : {} '.format(id,mytar))


  def clean_dir(self):
    folders_path = [os.path.join(self.pdfs_dir, fold).replace(' ', '\\ ') for fold in os.listdir(self.pdfs_dir)]
    for fold in folders_path:
      os.system('mv {}/* {}'.format(fold, self.pdfs_dir.replace(' ', '\\ ')))
      os.system('rm -r {}'.format(fold))



'''
Since the arxiv database could not be processed with papers class (in papers/raw.py), many methods are overwritten in
paper_ssrn and papers_ssrn.

[Documentation here can be more detailed if needed]
'''
from tqdm.notebook import tqdm
import pandas as pd
import numpy as np
from habanero import Crossref
import re
import ast
from bs4 import BeautifulSoup

from naimai.utils.regex import multiple_replace
from naimai.constants.regex import regex_remove_from_ssrn_fields, regex_journal_names, regex_journal_names2
from naimai.constants.paths import path_open_citations
from naimai.papers.raw import papers, paper_base
from naimai.decorators import update_naimai_dois
from naimai.utils.general import get_doi_by_title, get_soup

tqdm.pandas()
class paper_ssrn(paper_base):
    def __init__(self,df,idx_in_df):
        super().__init__()
        self.database = 'ssrn'
        self.file_name = idx_in_df
        self.paper_infos = df.iloc[idx_in_df,:]


    def get_doi(self) -> str:
        doi = self.paper_infos['doi']
        if doi:
            self.doi = doi
        else:
            self.doi = 'https://papers.ssrn.com/sol3/papers.cfm?abstract_id='+ str(self.paper_infos['abstract_id'])

    def get_fields(self) -> list:
        papers_field = self.paper_infos['papers_field']
        field = self.paper_infos['field']
        field_cleaned = re.sub(regex_remove_from_ssrn_fields,'',field).strip()
        self.fields = [field_cleaned,papers_field]

    def get_Abstract(self) -> str:
        self.Abstract = self.paper_infos['abstract_text'].replace('-\n', '').replace('\n', ' ').strip()

    def get_Title(self) -> str:
        self.Title = self.paper_infos['title'].replace('-\n', '').replace('\n', ' ').strip()

    def get_Authors(self) -> str:
        self.Authors = self.paper_infos['authors']

    def get_year(self) -> str:
        self.year = self.paper_infos['date']

    def get_journal(self) -> str:
        abstract_box = self.paper_infos['abstract_box']
        bs = BeautifulSoup(abstract_box, 'html.parser')
        journal = bs.find(name='div',attrs={'class': 'reference-info'})
        if journal:
            clean1 = re.sub(regex_journal_names,'',journal.text)
            self.Journal = re.sub(regex_journal_names2,'',clean1, flags=re.I).strip()
        else:
            self.Journal =  self.paper_infos['field']

    def replace_abbreviations(self):
        abbreviations_dict = self.get_abbreviations_dict()
        if abbreviations_dict:
            self.Abstract = multiple_replace(abbreviations_dict, self.Abstract)
            self.Title = multiple_replace(abbreviations_dict, self.Title)

    def get_numCitedBy(self) -> int:
        ssrn_page = 'https://papers.ssrn.com/sol3/papers.cfm?abstract_id'
        if ssrn_page not in self.doi:
            path = path_open_citations + self.doi
            soup = get_soup(path)
            soup_list = ast.literal_eval(soup.text)
            if isinstance(soup_list, list):
                self.numCitedBy = len(soup_list)

    def get_Keywords(self) -> str:
        keywords = self.paper_infos['keywords']
        if isinstance(keywords,str):
            self.Keywords = keywords.replace('Keywords: ', '')


class papers_ssrn(papers):
    def __init__(self, papers_path,data=None):
        super().__init__() # loading self.naimai_dois & other attributes
        if isinstance(data,pd.DataFrame):
            self.data = data[data['abstract_text'].notna()].reset_index(drop=True)
            self.data = self.data.fillna('')
            print('Len data : ', len(self.data))
        else:
            data2 = pd.read_csv(papers_path)
            self.data = data2[data2['abstract_text'].notna()].reset_index(drop=True)
            self.data['papers_field'] = papers_path.split('/')[-2].replace('_',' ')
            print('Len data : ', len(self.data))
            print('')
            print('Getting dois..')
            cr = Crossref()
            self.data['doi'] = self.data['title'].progress_apply(get_doi_by_title, args=(cr,))


    def add_paper(self,idx_in_data):
            new_paper = paper_ssrn(df=self.data,
                                    idx_in_df=idx_in_data)
            new_paper.get_doi()
            if not new_paper.is_in_database(self.naimai_dois):
                new_paper.get_Abstract()
                new_paper.get_fields()
                new_paper.get_Title()
                try:
                    new_paper.get_journal()
                except:
                    pass
                new_paper.get_Authors()
                new_paper.get_Keywords()
                new_paper.get_year()
                new_paper.replace_abbreviations()
                new_paper.get_numCitedBy()
                self.elements[new_paper.doi] = new_paper.save_dict()
                self.naimai_dois.append(new_paper.doi)


    @update_naimai_dois
    def get_papers(self,update_dois=False,idx_start=0,idx_finish=-1,show_tqdm=False):

        if show_tqdm:
            range_ = tqdm(self.data.iterrows(),total=len(self.data))
        else:
            range_= self.data.iterrows()
        for idx,_ in range_:
            self.add_paper(idx_in_data=idx)



'''
Attemps to extract data from the whole pmc papers (in xml). [Not Finished]

[Documentation here can be more detailed if needed]
'''

import ast
import re
import pandas as pd
from tqdm.notebook import tqdm
from spacy_langdetect import LanguageDetector
import spacy

from naimai.decorators import update_naimai_dois
from naimai.papers.raw import papers, paper_full_base
from naimai.constants.paths import path_open_citations
from naimai.constants.regex import regex_spaced_chars,regex_keywords
from naimai.constants.nlp import nlp_vocab
from naimai.utils.regex import multiple_replace
from naimai.utils.general import get_soup


class paper_pmc(paper_full_base):
    def __init__(self ,df ,idx_in_df):
        super().__init__()
        self.database ="pmc"
        self.paper_infos = df.iloc[idx_in_df,:]

    def get_doi(self):
        self.doi = self.paper_infos['doi']

    def get_Keywords_from_dict_abstract(self):
        '''
        get keywords from abstract when it's in a dict format & filter from abstract
        :return:
        '''
        for elt in self.Abstract:
            txt = self.Abstract[elt]
            keywords_in_txt = re.findall(regex_keywords, txt, re.I)
            if keywords_in_txt:
                self.Keywords = keywords_in_txt[0].strip()
                regex_filter = 'key\s?words:\s?' + self.Keywords
                self.Abstract[elt] = re.sub(regex_filter,'',txt,flags=re.I).strip()

    def get_Keywords_from_str_abstract(self):
        '''
        get keywords from abstract when it's in str format & filter from abstract
        :return:
        '''
        txt = self.Abstract
        keywords_in_txt = re.findall(regex_keywords, txt, re.I)
        if keywords_in_txt:
            self.Keywords = keywords_in_txt[0].strip()
            regex_filter = 'key\s?words:\s?' + self.Keywords
            self.Abstract = re.sub(regex_filter, '', txt, flags=re.I).strip()

    def get_Keywords(self):
        '''
        Get keywords from abstracts elements & remove them from abstract
        :return:
        '''
        if isinstance(self.Abstract,dict):
            self.get_Keywords_from_dict_abstract()
        elif isinstance(self.Abstract,str):
            self.get_Keywords_from_str_abstract()


    def get_Abstract(self,stacked=True,dict_format=False):
        '''
        clean & stack (if stacked=True) abstract elements into one text & get Keywords from abstract.
        The spaced abstract case (a b s t r a c t..) is considered. if stacked = False, it puts the abstract elements in
        str format. It can returns in dict format if dict_format=True
        :return:
        '''
        abstract_dict_str = self.paper_infos['abstract']
        abstract_dict = ast.literal_eval(abstract_dict_str)
        if 'text' not in abstract_dict.keys():
            abstract_dict = {elt: ' '.join(abstract_dict[elt]) for elt in abstract_dict}

        for elt in abstract_dict:
            clean1=re.sub('abstract', '',abstract_dict[elt], flags=re.I).strip()
            clean2 = self.clean_text(clean1)
            abstract_dict[elt]=clean2

        if stacked:
            abstract = ' '.join([elt for elt2 in list(abstract_dict.values()) for elt in elt2])
            no_space = re.findall('\w\w',abstract)
            if no_space: # normal case
                self.Abstract = abstract.replace('-\n', '').replace('\n', ' ')
            else: #spaced abstract, we "despace" 2 times
                despacing1 = re.sub(regex_spaced_chars, r'\1\2', abstract)
                self.Abstract = re.sub(regex_spaced_chars, r'\1\2', despacing1).replace('\n',' ')

        else:
            # result = {elt: ' '.join(abstract_dict[elt]) for elt in abstract_dict}
            if dict_format:
                self.Abstract = abstract_dict
            else:
                result_txt = ''
                for elt in abstract_dict:
                    result_txt+=' '+ elt+': '+abstract_dict[elt]
                self.Abstract= result_txt.strip()
        self.get_Keywords()
    def get_Title(self):
        self.Title = self.paper_infos['title'].replace('-\n', '').replace('\n', ' ')

    def get_Authors(self):
        authors_list_str = self.paper_infos['authors']
        authors_list = ast.literal_eval(authors_list_str)
        self.Authors = ', '.join(authors_list)

    def get_year(self):
        self.year = self.paper_infos['year']

    def get_journal(self):
        self.Journal =  self.paper_infos['journal']

    def get_body(self):
        body_dict_str = self.paper_infos['body']
        body_dict = ast.literal_eval(body_dict_str)
        return body_dict

    def headers2imr(self,headers: list) -> dict:
        '''
        get list of headers of the paper and convert it to imr sections
        :param headers:
        :return: {'abstract': [..], 'introduction': [..], 'method': [..], 'results':[..], 'rest': [..]}
        '''
        imr_sections_dict = {'abstract':[], 'introduction':[],'methods':[],'results':[],'references':[]}
        imr_sections_list = list(imr_sections_dict)
        imr_rgx = ['abstract', 'introduction', 'methods?', 'results?|discussion',
                       'references|funding|disclosures|acknowledgment']
        idx_section = 0
        for elt in headers:
            pointer = imr_sections_list[idx_section]
            nextp = imr_sections_list[idx_section + 1]
            try:
                nextp2 = imr_sections_list[idx_section + 2]
            except:
                pass
            nextp_rgx = imr_rgx[idx_section + 1]
            pattern_next = re.compile(nextp_rgx, flags=re.I)
            pattern_next2 = re.compile(nextp2, flags=re.I)
            if re.findall(pattern_next, elt):
                if idx_section + 2 < len(imr_sections_list):
                    imr_sections_dict[nextp] += [elt]
                    idx_section += 1
                else:
                    imr_sections_dict[nextp] += [elt]
            elif re.findall(pattern_next2, elt):
                imr_sections_dict[nextp2] += [elt]
                idx_section += 2
            else:
                imr_sections_dict[pointer] += [elt]
        return imr_sections_dict

    def get_abstract_extra(self, body, imr_sections):
        '''
        Clean & Add text before introduction to unclassified section
        :param body:
        :param imr_sections:
        :return:
        '''
        headers = imr_sections['abstract']
        if headers:
            for head in headers:
                text = ' '.join(body[head])
                cleaned_text = self.clean_text(text)
                self.unclassified_section[head] =cleaned_text

    def get_Introduction(self,body,imr_sections):
        '''
        get introductions elements & clean & stack them in introduction section
        :param body:
        :param imr_sections:
        :return:
        '''
        headers = imr_sections['introduction']
        if headers:
            for head in headers:
                text= ' '.join(body[head])
                cleaned_text = self.clean_text(text)
                self.Introduction[head] =cleaned_text

    def get_Methods(self,body,imr_sections):
        '''
        get methods elements & clean & stack them in method section
        :param body:
        :param imr_sections:
        :return:
        '''
        headers = imr_sections['methods']
        if headers:
            for head in headers:
                text = ' '.join(body[head])
                cleaned_text = self.clean_text(text)
                self.Methods[head]= cleaned_text

    def get_Results(self, body, imr_sections):
        '''
        get methods elements & clean & stack them in method section
        :param body:
        :param imr_sections:
        :return:
        '''
        headers = imr_sections['results']
        if headers:
            for head in headers:
                text = ' '.join(body[head])
                cleaned_text = self.clean_text(text)
                self.Results[head] = cleaned_text

    def get_non_imrad_content(self,body):
        '''
        when the format is not imrad, we gather everything in unclassified section
        :param body:
        :return:
        '''
        headers = list(body)
        for head in headers:
            text = ' '.join(body[head])
            cleaned_text = self.clean_text(text)
            self.unclassified_section[head] = cleaned_text

    def get_content(self,stacked_abstract=True,abstract_dict_format=False):
        '''
        get body infos and decompose it to abstract and IMR sections
        :return:
        '''
        body = self.get_body()
        headers = list(body.keys())
        self.get_Abstract(stacked=stacked_abstract,dict_format=abstract_dict_format)
        try:
            imr_sections = self.headers2imr(headers)
            self.get_abstract_extra(body=body, imr_sections=imr_sections)
            self.get_Introduction(body=body,imr_sections=imr_sections)
            self.get_Methods(body=body,imr_sections=imr_sections)
            self.get_Results(body=body,imr_sections=imr_sections)
        except: #not IMRAD format
            self.get_non_imrad_content(body)



    def get_numCitedBy(self):
        if self.doi!=self.file_name:
            path = path_open_citations + self.doi
            soup = get_soup(path)
            soup_list = ast.literal_eval(soup.text)
            if isinstance(soup_list,list):
                self.numCitedBy = len(soup_list)

    def is_paper_english(self,nlp) -> bool:
        '''
        Detect language if paper language is english based on its title.
        :return:
        '''
        if self.Title:
            title_nlp = nlp(self.Title)
            language_score=title_nlp._.language
            condition_english = (language_score['language']=='en') and (language_score['score']>0.7)
            if condition_english:
                return True
        lang = language_score['language']
        score = language_score['score']
        print(f'Langage {lang} - Score {score}')
        return False


    def replace_abbreviations(self):
        abbreviations_dict = self.get_abbreviations_dict()
        if abbreviations_dict:
            if isinstance(self.Abstract,str):
                self.Abstract = multiple_replace(abbreviations_dict, self.Abstract)
            elif isinstance(self.Abstract,dict):
                for elt in self.Abstract:
                    self.Abstract[elt] = multiple_replace(abbreviations_dict, self.Abstract[elt])
            for elt in self.Introduction:
                self.Introduction[elt] = multiple_replace(abbreviations_dict, self.Introduction[elt])
            for elt in self.Methods:
                self.Methods[elt] = multiple_replace(abbreviations_dict, self.Methods[elt])
            for elt in self.Results:
                self.Results[elt] = multiple_replace(abbreviations_dict, self.Results[elt])
            self.Title = multiple_replace(abbreviations_dict, self.Title)

class papers_pmc(papers):
    def __init__(self, papers_path,stacked_abstract=True,abstract_dict_format=False,nlp=None):
        super().__init__() # loading self.naimai_dois & other attributes
        self.naimai_dois = []
        self.data = pd.read_csv(papers_path)
        self.stacked_abstract=stacked_abstract
        self.abstract_dict_format=abstract_dict_format
        print('Len data : ', len(self.data))
        if nlp:
            self.nlp = nlp
        else:
            print('Loading nlp vocab..')
            self.nlp = spacy.load(nlp_vocab)
            self.nlp.add_pipe(LanguageDetector(), name='language_detector', last=True)

    def add_paper(self,idx_in_data):
            new_paper = paper_pmc(df=self.data,
                                    idx_in_df=idx_in_data)
            new_paper.get_doi()
            if not new_paper.is_in_database(self.naimai_dois):
                self.naimai_dois.append(new_paper.doi)
                new_paper.get_Title()
                if new_paper.is_paper_english(self.nlp):
                    new_paper.get_Authors()
                    new_paper.get_journal()
                    new_paper.get_year()
                    new_paper.get_content(stacked_abstract=self.stacked_abstract,
                                          abstract_dict_format=self.abstract_dict_format)
                    new_paper.replace_abbreviations()
                    new_paper.get_numCitedBy()
                    self.elements[new_paper.doi] = new_paper.save_dict()
                else:
                    print(f'in Title : {new_paper.Title} - index in data {idx_in_data} is not english..')


    @update_naimai_dois
    def get_papers(self):
        for idx,_ in tqdm(self.data.iterrows(),total=len(self.data)):
            self.add_paper(idx_in_data=idx)


'''
Modified grobid client from Grobid source: https://github.com/kermitt2/grobid. The GrobidClient object
takes a directory of PDFs and converts it to dictionaries.
'''

import os
import json
import time
import concurrent.futures
import requests

from .client import ApiClient


class ServerUnavailableException(Exception):
    pass


class GrobidClient(ApiClient):

    def __init__(self, url,
                 coordinates=["persName", "figure", "ref", "biblStruct", "formula", "s"],
                 sleep_time=5,
                 timeout=180,
                 config_path=None,
                 check_server=True):
        self.config = {
            'url': url,
            'coordinates': coordinates,
            'sleep_time': sleep_time,
            'timeout': timeout
        }
        if config_path:
            self._load_config(config_path)
        if check_server:
            self._test_server_connection()

    def _load_config(self, path="./config.json"):
        """
        Load the json configuration
        """
        config_json = open(path).read()
        self.config = json.loads(config_json)

    def _test_server_connection(self):
        """Test if the server is up and running."""
        the_url = self.config['url']
        r = requests.get(the_url)

        status = r.status_code

        if status != 200:
            print("GROBID server does not appear up and running " + str(status))
        else:
            print("GROBID server is up and running")

    def process(
            self,
            service,
            input_path,
            n=10,
            generateIDs=False,
            consolidate_header=True,
            consolidate_citations=False,
            include_raw_citations=False,
            include_raw_affiliations=False,
            tei_coordinates=False,
            segment_sentences=False,
            verbose=False,
            path_export='',
    ):
        '''
        Process a pdf files and returns the results of the binary data (pdf read files in dictionary format) is given.
        Other params are explained in the grobid documentation.

        :param binary_data: when read PDF files are read in binary format
        :return:
        '''
        input_files = {}
        result = {}
        filenames = sorted(list(input_path))
        for filename in filenames:
            if filename.endswith(".pdf") or filename.endswith(".PDF") or \
                    (service == 'processCitationList' and (filename.endswith(".txt") or filename.endswith(".TXT"))):
                if verbose:
                    try:
                        print(filename)
                    except Exception:
                        # may happen on linux see https://stackoverflow.com/questions/27366479/python-3-os-walk-file-paths-unicodeencodeerror-utf-8-codec-cant-encode-s
                        pass
                input_files[filename]=input_path[filename]

                result[filename] = self.process_batch(
                    service,
                    input_files,
                    n,
                    generateIDs,
                    consolidate_header,
                    consolidate_citations,
                    include_raw_citations,
                    include_raw_affiliations,
                    tei_coordinates,
                    segment_sentences,
                    verbose,
                )
                input_files = {}

        if path_export:
            print('Files exported in :', path_export)
            for fname in result:
                text = result[fname]
                output_fname = os.path.join(path_export,fname.replace('pdf','xml'))
                with open(output_fname,'w',encoding='utf8') as xml_file:
                    xml_file.write(text)

        return result

    def process_batch(
            self,
            service,
            input_files,
            n,
            generateIDs,
            consolidate_header,
            consolidate_citations,
            include_raw_citations,
            include_raw_affiliations,
            tei_coordinates,
            segment_sentences,
            verbose=False,
    ):
        if verbose:
            print(len(input_files), "files to process in current batch")

        # we use ThreadPoolExecutor and not ProcessPoolExecutor because it is an I/O intensive process
        with concurrent.futures.ThreadPoolExecutor(max_workers=n) as executor:
            # with concurrent.futures.ProcessPoolExecutor(max_workers=n) as executor:
            results = []
            for input_file in input_files:
                selected_process = self.process_pdf
                if service == 'processCitationList':
                    selected_process = self.process_txt

                r = executor.submit(
                    selected_process,
                    service,
                    input_file,
                    input_files[input_file],
                    generateIDs,
                    consolidate_header,
                    consolidate_citations,
                    include_raw_citations,
                    include_raw_affiliations,
                    tei_coordinates,
                    segment_sentences)

                results.append(r)
        for r in concurrent.futures.as_completed(results):
            input_file, status, text = r.result()

            if text is None:
                print("Processing of", input_file, "failed with error", str(status))
                return ''
            else:
                return text

    def process_pdf(
            self,
            service,
            filename,
            pdf_file,
            generateIDs,
            consolidate_header,
            consolidate_citations,
            include_raw_citations,
            include_raw_affiliations,
            tei_coordinates,
            segment_sentences
    ):
        files = {
            "input": (
                filename,
                pdf_file,
                "application/pdf",
                {"Expires": "0"},
            )
        }
        the_url = self.config['url']
        the_url += "/api/" + service

        # set the GROBID parameters
        the_data = {}
        if generateIDs:
            the_data["generateIDs"] = "1"
        if consolidate_header:
            the_data["consolidateHeader"] = "1"
        if consolidate_citations:
            the_data["consolidateCitations"] = "1"
        if include_raw_citations:
            the_data["includeRawCitations"] = "1"
        if include_raw_affiliations:
            the_data["includeRawAffiliations"] = "1"
        if tei_coordinates:
            the_data["teiCoordinates"] = self.config["coordinates"]
        if segment_sentences:
            the_data["segmentSentences"] = "1"

        try:

            res, status = self.post(
                url=the_url, files=files, data=the_data, headers={"Accept": "text/plain"},
                timeout=self.config['timeout']
            )

            if status == 503:
                time.sleep(self.config["sleep_time"])
                return self.process_pdf(
                    service,
                    filename,
                    pdf_file,
                    generateIDs,
                    consolidate_header,
                    consolidate_citations,
                    include_raw_citations,
                    include_raw_affiliations,
                    tei_coordinates,
                    segment_sentences
                )
        except requests.exceptions.ReadTimeout:
            return (pdf_file, 408, None)

        return (pdf_file, status, res.text)

    def process_txt(
            self,
            service,
            txt_file,
            generateIDs,
            consolidate_header,
            consolidate_citations,
            include_raw_citations,
            include_raw_affiliations,
            tei_coordinates,
            segment_sentences
    ):
        # create request based on file content
        references = None
        with open(txt_file) as f:
            references = [line.rstrip() for line in f]

        the_url = self.config['url']
        the_url += "/api/" + service

        # set the GROBID parameters
        the_data = {}
        if consolidate_citations:
            the_data["consolidateCitations"] = "1"
        if include_raw_citations:
            the_data["includeRawCitations"] = "1"
        the_data["citations"] = references
        res, status = self.post(
            url=the_url, data=the_data, headers={"Accept": "application/xml"}
        )

        if status == 503:
            time.sleep(self.config["sleep_time"])
            return self.process_txt(
                service,
                txt_file,
                generateIDs,
                consolidate_header,
                consolidate_citations,
                include_raw_citations,
                include_raw_affiliations,
                tei_coordinates,
                segment_sentences
            )

        return (txt_file, status, res.text)


regex_email = r'\w+@\w+\.\w+'
regex_url = r'http([^\s]+)'
regex_words=r'[^a-zA-Z+]'
regex_words_authors=r'[^a-zA-Z+\.,\-;]'
regex_words_commas=r'[^a-zA-Z(,|.) ]'
regex_not_converted = r'[\x00-\x08\x0b\x0c\x0e-\x1f\x7f-\xff]'
regex_not_converted2 = r'\u2009|\xa0|\u202f'
regex_abstract1 = r'a\s*b\s*s\s*t\s*r\s*a\s*c\s*t'
regex_abstract2 = r's\s*u\s*m\s*m\s*a\s*r\s*y'
regex_year = '[12][0-9]{3}a?b?c?'
regex_sentences = r'\.(?=\s+[A-Z])'
regex_words_numbers_some=r'[^A-Za-z0-9\s\(\)\.,;\-]'
regex_etal = 'et\s+al\.\s?'
regex_etal_and = 'et\s+al\.,?\s+|and'
regex_etal_and_plus=regex_etal_and+'|\(|\)'
regex_remove_nbs_from_end = '[^0-9]+$'
regex_capital=r'\b(?=[A-Z])'
regex_cid = r'\(?cid:\d+\)?'
regex_arxiv_filename = '\d{4}\.\d+\.pdf'
arxiv_pdfs_url = 'https://arxiv.org/pdf/'
regex_words_in_brackets = '\((\w.+?)\)'
regex_in_brackets='\((.*?)\)'
regex_and_operators = ' AND |&|\+'
regex_or_operators = '\||OR'
regex_exact_match='"(.*?)"'

study_terms = 'paper|study|investigation|work|research'
verbs_terms =  'hypothesize|aim(?:ed)?|proposed?|remedy|discuss(?:ed)?|evaluated?|used?|explored?|describes?|developed?|introduced?|present(?:ed)?|investigated?|examined?|show(?:ed)?'
objective_terms = ' objectives? | purposes? | goals? | propose | aims | examines? | investigate '
rgx1 = '[^.]*(?:this|our|present) (?:'+study_terms+')[^.]*\.'
rgx2 = '[^.]*(?:we|authors|is|were|are|have|has|was) (?:' +verbs_terms +')[^.]*\.'
rgx3 = '[^.]* (?:' + objective_terms+')[^.]*\.?'
rgx4 = '[^.]* here[^.]*\.'
rgx5 = '[^.]*to this end[^.]*\.'
rgx6 = '[^.]*the present[^.]*\.'
rgx7 = '(?:to \w+ this, )(?:.*?)\.'
rgx_or = '|'
regex_objectives = rgx1 + rgx_or + rgx2 + rgx_or + rgx3 + rgx_or + rgx4 + rgx_or + rgx5 + rgx_or + rgx6 + rgx_or + rgx7

regex_spaced_chars = '(.) (.)'
regex_abbrvs = 'i\.e\.?,?|e\.g\.?,?'
regex_abbrvs2 = 'cf\.'
regex_some_brackets = r'\((figure|table|meaning).*?\)'
regex_numbers_in_brackets = r'(\(|\[)\d\d?(\)|\])\s?'
regex_eft = r'(\(|\[)?(figure|fig\.|table|equation|eqs?\.)(\)|\])?'
regex_explained_2pts = '(:\s?.*?)\.'
regex_equation_tochange=r'(\w+)\s?=\s?(\w+(?:\.\w+))'
regex_equation='\w+\s?=\s?\w+(?:\.\w+)?'
regex_keywords='key\s?words:(.*)'
regex_background = 'introduction|background'
regex_objective = 'purpose|aim|objective'
regex_methods = 'method|materials|approach'
regex_results = 'result|finding|conclusion'
regex_doi = r"""
    ((\(?doi(\s)*\)?:?(\s)*)       # 'doi:' or 'doi' or '(doi)' (upper or lower case)
    |(https?://(dx\.)?doi\.org\/))?         # or 'http://(dx.)doi.org/'  (neither has to be present)
    (?P<doi>10\.                            # 10.                        (mandatory for DOI's)
    \d{3,7}                                 # [0-9] x 3-7
    (\.\w+)*                                # subdivisions separated by . (doesn't have to be present)
    (/|%2f)                                 # / (possibly urlencoded)
    [\w\-_:;\(\)/\.<>]+                     # any character
    [\w\-_:;\(\)/<>])                       # any character excluding a full stop
    """
regex_references = r'references(.*)\b'
regex_urlsoup=r'q=(.*)&sa=U&ved'
regex_paper_year = '19[0-9]{2}|20[01][0-9]|202[0-2]'

regex_filtered_words_obj = 'this section|Corresponding| PhD '
regex_remove_from_ssrn_fields = '&\s?|eJournal|Educator:?|Courses, Cases & Teaching|Teaching'
regex_journal_names = '[^A-Za-z\s]'
regex_journal_names2 = 'volume|issue|\n|doi'

regex_abstract_omr = 'objectives?:|results?:|questions?:|purposes?:|findings:|methods:|conclusion:|report:|discussions?:'
regex_review = 'review|summarizes?|document'


'''
Since the arxiv database could not be processed with papers class (in papers/raw.py), many methods are overwritten in
paper_arxiv and papers_arxiv.

[Documentation here can be more detailed if needed]
'''
from tqdm.notebook import tqdm
import dask.bag as db
import json
import re
import ast

from naimai.utils.regex import multiple_replace
from naimai.utils.general import get_soup
from naimai.papers.raw import papers, paper_base
from naimai.constants.fields import arxiv_fields_abbrevs, arxiv_fields_categories
from naimai.constants.regex import regex_year
from naimai.constants.paths import path_open_citations
from naimai.decorators import update_naimai_dois


class paper_arxiv(paper_base):
    def __init__(self,arxiv_id,metadata_df,idx_in_metadata_df,category):
        super().__init__()
        self.database = 'arxiv'
        self.file_name = arxiv_id
        # self.metadata_df = metadata_df
        self.category = category
        self.idx = idx_in_metadata_df
        self.paper_infos = metadata_df.iloc[idx_in_metadata_df,:]

    def get_doi(self):
        doi = self.paper_infos['doi']
        if not isinstance(doi,str):
            doi = self.file_name
        self.doi = doi


    def get_fields(self):
        field = self.category.split('.')[0]
        fields  = arxiv_fields_abbrevs[field] + ', ' + arxiv_fields_categories[field][self.category]
        self.fields = fields.replace('-', ',').split(',')

    def get_Abstract(self):
        self.Abstract = self.paper_infos['abstract'].replace('-\n', '').replace('\n', ' ')

    def get_Title(self):
        self.Title = self.paper_infos['title'].replace('-\n', '').replace('\n', ' ')

    def get_Authors(self):
        self.Authors = ', '.join([' '.join(at[::-1]) for at in self.paper_infos['authors_parsed']]).strip()

    def get_year(self):
        #self.year = year_from_arxiv_fname(self.file_name)
        self.year = self.paper_infos['year']

    def get_journal(self):
        self.Journal =  self.paper_infos['journal-ref']

    def replace_abbreviations(self):
        abbreviations_dict = self.get_abbreviations_dict()
        if abbreviations_dict:
            self.Abstract = multiple_replace(abbreviations_dict, self.Abstract)
            self.Title = multiple_replace(abbreviations_dict, self.Title)

    def get_numCitedBy(self):
        if self.doi!=self.file_name:
            path = path_open_citations + self.doi
            soup = get_soup(path)
            soup_list = ast.literal_eval(soup.text)
            if isinstance(soup_list,list):
                self.numCitedBy = len(soup_list)


class papers_arxiv(papers):
    def __init__(self, arxiv_metadata_dir,category):
        super().__init__() # loading self.naimai_dois & other attributes
        self.arxiv_metadata_dir = arxiv_metadata_dir
        self.category = category
        self.metadata_df = None
        self.abstracts = []
        self.titles = []
        self.authors = []
        self.files_ids = []

    def get_infos(self):
        all_docs = db.read_text(self.arxiv_metadata_dir).map(json.loads)
        docs = all_docs.filter(lambda x: x['categories'] == self.category)
        filtered_docs = docs.filter(lambda x: x['comments'] != 'This paper has been withdrawn')
        print('>> Getting metadata_df..')
        self.metadata_df = filtered_docs.to_dataframe()[['id', 'authors_parsed', 'title', 'abstract', 'doi','journal-ref','versions']].compute()
        self.metadata_df['year'] = self.metadata_df['versions'].apply(lambda x: re.findall(regex_year, x[0]['created'])[0])
        self.files_ids = self.metadata_df['id']

    def add_paper(self,arxiv_id,idx_in_metadata_df,force_naimai_dois=False):
            new_paper = paper_arxiv(arxiv_id=arxiv_id,
                                    metadata_df=self.metadata_df,
                                    idx_in_metadata_df=idx_in_metadata_df,
                                    category=self.category)
            new_paper.get_doi()
            if force_naimai_dois:
                condition=1
            else:
                condition = not new_paper.is_in_database(self.naimai_dois) #the doi doesn't exist in the database
            if condition:
                new_paper.get_Abstract()
                new_paper.get_fields()
                new_paper.get_Title()
                new_paper.get_journal()
                new_paper.get_Authors()
                new_paper.get_year()
                new_paper.replace_abbreviations()
                try:
                    new_paper.get_numCitedBy()
                except:
                    print('problem of citing in paper ', new_paper.doi)
                self.elements[arxiv_id] = new_paper.save_dict()
                self.naimai_dois.append(new_paper.doi)


    @update_naimai_dois
    def get_papers(self,update_dois=False,force_naimai_dois=False,idx_start=0,idx_finish=-1):
        self.get_infos()
        files_ids=self.files_ids[idx_start:idx_finish]
        for idx_in_metadata_df,arxiv_id in tqdm(enumerate(files_ids), total=len(files_ids)):
            self.add_paper(arxiv_id=arxiv_id,idx_in_metadata_df=idx_in_metadata_df,force_naimai_dois=force_naimai_dois)

        print('Objs problem exported in objectives_pbs.txt')


from .base import BaseQuerier
from naimai.utils.regex import lemmatize_query


class KeywordsQuerier(BaseQuerier):
    def __init__(self,field):
        super().__init__(field=field, is_custom=False)

    def get_papers_with_exact_match(self,query: str, year_from=0, year_to=3000,top_n=5) -> tuple:
        '''
        find papers for a query with exact match
        '''
        selected_papers = self.sql_manager.search_with_exact_match(query,year_from=year_from,year_to=year_to,top_n=top_n)
        selected_papers_fnames = list(selected_papers)
        return selected_papers, selected_papers_fnames


    def get_papers_with_OR_operator(self,query: str, year_from=0, year_to=3000,top_n=5) -> tuple:
        '''
        find papers for a query with or operator
        '''

        selected_papers = self.sql_manager.search_with_OR_operator(query,year_from=year_from,year_to=year_to,top_n=top_n)
        selected_papers_fnames = list(selected_papers)
        return selected_papers, selected_papers_fnames


    def get_papers_with_AND_operator(self,query: str, year_from=0, year_to=3000,top_n=5) -> tuple:
        '''
        find papers for a query with and operator
        '''

        selected_papers = self.sql_manager.search_with_AND_operator(query,year_from=year_from,year_to=year_to,top_n=top_n)
        selected_papers_fnames = list(selected_papers)
        return selected_papers, selected_papers_fnames

    def get_papers_for_tfidf_semantics(self,query: str, year_from=0, year_to=3000,top_n=5) -> tuple:
        '''
        find papers using tf idf
        :param query:
        :return:
        '''

        lemmatized_query = lemmatize_query(self.nlp,query)
        selected_papers = self.sql_manager.get_by_lemmatized_query(lemmatized_query=lemmatized_query, year_from=year_from, year_to=year_to, top_n=top_n)
        selected_papers_fnames = list(selected_papers)
        return selected_papers, selected_papers_fnames





from tqdm.notebook import tqdm
from naimai.models.text_generation.query_generation import QueryGeneration
import torch
import pandas as pd
import numpy as np
import random
from sentence_transformers import SentenceTransformer, InputExample, losses, models, datasets, evaluation


class Search_Model:
  def __init__(self, field: str,papers: dict,eval_papers: dict,batch_size=16,n_epochs=10,checkpoint='sentence-transformers/msmarco-distilbert-base-dot-prod-v3'):
    self.field = field
    self.batch_size = batch_size
    self.n_epochs = n_epochs
    self.checkpoint = checkpoint
    self.training_data_df = None
    self.eval_data_df= None
    self.model = None
    self.papers = papers
    self.eval_papers= eval_papers

  def prepare_data(self,papers):
    keys = list(papers.keys())
    messages = [papers[fname]['messages'] for fname in keys]
    messages = [elt for elt2 in messages for elt in elt2]
    titles = [papers[fname]['title'] for fname in keys if 'title' in papers[fname]]

    sentences = messages + titles
    qgen = QueryGeneration(nb_queries=2)
    data_dict = {'sentences': [], 'queries': []}

    for sentence in tqdm(sentences):
      if len(sentence.split()) > 5:
        queries = qgen.from_message(sentence)
        if queries:
          for qry in queries:
            data_dict['sentences'].append(sentence)
            data_dict['queries'].append(qry)
    return pd.DataFrame(data_dict)

  def process_data(self):
    query, sentences = self.training_data_df['queries'].to_list(), self.training_data_df['sentences'].to_list()
    train_examples = [InputExample(texts=[qry, parag]) for qry, parag in zip(query, sentences)]
    random.shuffle(train_examples)
    self.processed_data = datasets.NoDuplicatesDataLoader(train_examples, batch_size=self.batch_size)

  def create_model(self):
    word_emb = models.Transformer(self.checkpoint)
    pooling = models.Pooling(word_emb.get_word_embedding_dimension())
    if torch.cuda.is_available():
      print('  >> GPU Used in search model !')
      self.model = SentenceTransformer(modules=[word_emb, pooling], device='cuda')
    else:
      print('  >> No GPU used in search model..')
      self.model = SentenceTransformer(modules=[word_emb, pooling])

  def train(self):
    # sentences = self.eval_data_df['sentences']
    # queries = self.eval_data_df['queries']
    # scores= [np.random.uniform(low=0.8, high=1.0) for _ in range(len(queries))]
    # evaluator = evaluation.EmbeddingSimilarityEvaluator(sentences, queries, scores)
    train_loss = losses.MultipleNegativesRankingLoss(self.model)
    warmup_steps = int(len(self.processed_data) * self.n_epochs * 0.1)
    self.model.fit(train_objectives=[(self.processed_data, train_loss)], epochs=self.n_epochs,
                    warmup_steps=warmup_steps, show_progress_bar=True)

  def fine_tune(self,model_path_saving=''):
    print('>> Preparing training data..')
    self.training_data_df = self.prepare_data(self.papers)
    # print('>> Preparing eval data..')
    # self.eval_data_df = self.prepare_data(self.eval_papers)
    self.process_data()

    print('>> Modelling..')

    if not self.model:
      self.create_model()
    else:
      print('>> Training the loaded encoding model  .. ')
      self.model.to('cuda')
    self.train()

    if model_path_saving:
      print('>> Model saving..')
      self.model.save(model_path_saving)
    print('Done!')



from bs4 import BeautifulSoup
import re
from tqdm.notebook import tqdm
from Bio import Entrez
import pandas as pd

def download_papers(start,n_iter,delta,search_results):
    for iter in tqdm(range(1, n_iter + 1)):
        handle = Entrez.efetch(db="pmc", rettype="full", retmode="xml", retstart=start, retmax=delta,
                               webenv=search_results["WebEnv"], query_key=search_results["QueryKey"])
        res = handle.read()

        fname = 'file.xml'
        with open(fname, 'wb') as f:
            f.write(res)

        dbase = 'pmc'
        crawler = Pubmed_Crawler(xml_filename=fname, database=dbase)
        crawler.get_docs()

        df = pd.DataFrame(crawler.docs)
        if df.shape[0]<5:
            print('>>>>>> WARNING !!!!!!!')
        fname = str(start) + '_' + str(start + delta) + '.csv'
        print('fname : ', fname)
        path_csv = "drive/MyDrive/MyProject/data_pipeline2/landing_zone/Pubmed/" + fname
        df.to_csv(path_csv)

        start = start + delta


def extract_author_name(contrib):
    try:
        lname = contrib.find('surname').text
        fname = contrib.find('given-names').text
        return fname + ' ' + lname
    except:
        return ''

class Pubmed_Crawler:
    def __init__(self, xml_filename, database):
        self.xml_filename = xml_filename
        self.xml_soup = None
        self.database = database
        self.articles = []
        self.docs = {'title': [], 'authors': [], 'year': [], "abstract": [], "doi": [],
                     "database": [], 'journal': [], 'body': []}

    def read_file(self):
        xml_file = open(self.xml_filename, 'r')
        self.xml_soup = BeautifulSoup(xml_file, "lxml")

    def get_articles(self):
        self.articles= self.xml_soup.find_all('article')

    def get_doi(self, article):
        doi = article.find(name='article-id', attrs={'pub-id-type': 'doi'})
        if doi:
            return doi.text
        return

    def get_title(self, article):
        title = article.find('article-title')
        if title:
            return title.text
        return

    def get_year(self, article):
        year = article.find('year')
        if year:
            return year.text
        return

    def get_authors(self, article):
        authors_contribs = article.find_all(name='contrib', attrs={'contrib-type': 'author'})
        return [extract_author_name(contrib) for contrib in authors_contribs]

    def get_journal(self, article):
        journal = article.find('journal-title')
        if journal:
            return journal.text
        return

    def get_abstract(self, article):
        abstract_soup = article.find('abstract')
        if abstract_soup:
            abstract_elements = abstract_soup.find_all('sec')
            if abstract_elements :
              try:
                  return {elt.find('title').text: [elt.text.strip() for elt in elt.find_all('p')] for elt in abstract_elements}
              except:
                  return {'text': abstract_soup.text}
            else:
              return {'text': abstract_soup.text}
        return


    def get_body_clean(self, article,abstract):
        body_elements = article.find_all(name='sec', id=re.compile('\w+'))
        res = {}
        for elt in body_elements:
          try:
            title = elt.find('title').text
            text=[elt.text.strip() for elt in elt.find_all('p')]
            if title not in abstract.keys():
              res[title] = text
          except:
            pass
        return res

    def get_body_greedy(self,article,abstract):
        abstract_phrases = []
        for stc in abstract.values():
            abstract_phrases+=stc

        all_phrases = [elt.text.strip() for elt in article.find_all('p')]
        body_phrases = [elt for elt in all_phrases if elt not in abstract_phrases][:-5]
        return body_phrases

    def get_body(self,article,abstract):
        result = self.get_body_clean(article,abstract)
        if result:
            return result
        result = self.get_body_greedy(article,abstract)
        return {'text': result}

    def get_docs(self):
        self.read_file()
        self.get_articles()
        for article in self.articles:
            abstract = self.get_abstract(article)
            if abstract:
                self.docs['abstract'].append(abstract)
                self.docs['title'].append(self.get_title(article))
                self.docs['authors'].append(self.get_authors(article))
                self.docs['year'].append(self.get_year(article))
                self.docs['doi'].append(self.get_doi(article))
                self.docs['database'].append(self.database)
                self.docs['journal'].append(self.get_journal(article))
                self.docs['body'].append(self.get_body(article,abstract))


from unittest import TestCase
from naimai.papers.only_abstracts.issn import papers_issn
from naimai.pipelines.producers import Custom_Producer
from naimai.data_queriers.custom_querier import CustomQuerier
import os


class CustomQuerierTest(TestCase):
    """
    Test custom querier, by producing issn papers in test input data.
    """

    @classmethod
    def setUpClass(cls) -> None:
        cls.review = """
        - Haruna and Nkongolo (2020) evaluated the influence of three years of cover crop, tillage, and crop rotation on selected soil nutrients.\n- Fontanazza et al. (2021) accelerated the degradation of biodegradable plastic (BP) mulch films and decreased the plastic pollution in agriculture.\n- Zarzecka et al. (2020) observed in plots where potato had been treated with herbicides and herbicides mixed wit
        """
        cls.references = """
        - Samuel I. Haruna, N. Nkongolo, 2020, Influence of Cover Crop, Tillage, and Crop Rotation Management on Soil Nutrients, Agriculture.\n- Stefania Fontanazza, A. Restuccia, G. Mauromicale, A. Scavo, C. Abbate, 2021, Pseudomonas putida Isolation and Quantification by Real-Time PCR in Agricultural Soil Biodegradable Mulching, Agriculture.\n- K. Zarzecka, M. Gugała, A. Sikorska, K. Grzywacz, M. Niewęgło
        """

        cls.first_paper_result = {
            "website": "",
            "year": 2020,
            "database": "issn",
            "messages": [
                "Cover cropping, tillage and crop rotation management can influence soil nutrient availability and crop yield through changes in soil physical, chemical and biological processes.",
                "The objective of this study was to evaluate the influence of three years of cover crop, tillage, and crop rotation on selected soil nutrients.",
                "Twenty-four plots each of corn (Zea mays) and soybean (Glycine max) were established on a 4.05 ha field and arranged in a three-factor factorial design.",
                "The three factors (treatments) were two methods of tillage (no-tillage (NT) vs. moldboard plow [conventional] tillage (CT)), two types of cover crop (no cover crop (NC) vs. cover crop (CC)) and four typess of rotation (continuous corn, continuous soybean, corn/soybean and soybean/corn).",
            ],
            "reported": "Haruna and Nkongolo (2020) evaluated the influence of three years of cover crop, tillage, and crop rotation on selected soil nutrients.",
            "title": "Influence of Cover Crop, Tillage, and Crop Rotation Management on Soil Nutrients",
            "numCitedBy": 5.0,
            "journal": "Agriculture",
            "authors": "Haruna and Nkongolo (2020)",
            "allauthors": "Samuel I. Haruna, N. Nkongolo",
        }
        # reading data
        path = os.path.join("tests", "papers", "input_data", "issn_input.csv")
        obj = papers_issn(path, database="issn")
        obj.get_papers(update_dois=False, show_tqdm=False, check_database=False)

        # producing data
        producer = Custom_Producer(papers_dict=obj.elements)
        producer.produce_custom_papers(show_tqdm=False)

        # querying
        produced_papers = producer.produced_custom_papers
        cls.querier = CustomQuerier(produced_papers=produced_papers)
        cls.query = "cover crop"

        # reviewing
        cls.review_result, cls.references_result = cls.querier.review(cls.query)

    def test_papers_results(self):
        """
        test numcitedby of the first paper
        :return:
        """
        papers = self.querier.find_papers(self.query)
        first_paper = papers[0]
        self.assertIsInstance(papers,list)
        self.assertEqual(first_paper, self.first_paper_result)

    def test_review(self):
        """
        test if correct fnames are read
        :return:
        """
        self.assertIn(self.review_result[:400], self.review)

    def test_references(self):
        """
        test title of the first paper
        :return:
        """
        self.assertIn(self.references_result[:400], self.references)



from tqdm.notebook import tqdm
import os
from bs4 import BeautifulSoup
import random
import requests
import time

class Scholar_Crawler:
    def __init__(self, person,field, path,t_min=3,t_max=5):
        self.field = field
        self.person = person
        self.path = path
        self.soup = self.get_local_soup(path)
        self.docs = {'title': [], 'abstract': [], 'authors': [], 'date': [], 'field': [],'website':[], 'numCitedBy': [],'journal': []}
        self.t_min=t_min
        self.t_max=t_max

    def get_soup(self, path):
        header = {}
        header[
            'User-Agent'] = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'
        soup = BeautifulSoup(requests.get(path, headers=header, timeout=15).content, 'html.parser')
        slp=random.randint(self.t_min,self.t_max)
        time.sleep(slp)
        return soup

    def get_local_soup(self,path):
        with open(path) as fp:
            soup = BeautifulSoup(fp, 'html.parser')
        return soup

    def get_abstract_text(self,abstract_box):
        return abstract_box.find_all(name='div', attrs={"class": "abstract-text"})[0].find('p').getText()

    def get_years(self):
        return [elt.text for elt in self.soup.find_all(name='span', attrs={'class': 'gsc_a_h gsc_a_hc gs_ibl'})]

    def get_titles_elts(self):
        return self.soup.find_all(name='a', attrs={'class': 'gsc_a_at'})

    def get_articles_pages(self,titles_elts):
        return [elt['href'] for elt in titles_elts]

    def get_titles(self, titles_elts):
        return [elt.text for elt in titles_elts]

    def get_numCitedBy(self):
        return [elt.text for elt in self.soup.find_all(name='a', attrs={'class': 'gsc_a_ac gs_ibl'})]

    def get_article_page(self, website, dir_file, title,idx_start,idx):
        title = title + '.html'
        path_file = os.path.join(dir_file, title)

        if idx<idx_start:
            soup = self.get_local_soup(path_file)
        else:
            soup = self.get_soup(website)

        with open(path_file, "w") as file:
            file.write(str(soup))
        return path_file

    def get_article_metadata(self,soup):
        return soup.find_all(name='div', attrs={'class': 'gsc_oci_value'})

    def get_abstract(self,soup):
        abstract =soup.find_all(name='div', attrs={'class': 'gsh_csp'})
        if abstract:
            return abstract[0].text
        else:
            return ''

    def get_docs(self,idx_start=0):
        print('>> Local data..')
        titles_elts= self.get_titles_elts()
        self.docs['website'] = self.get_articles_pages(titles_elts)
        self.docs['title'] = self.get_titles(titles_elts)
        self.docs['numCitedBy'] = self.get_numCitedBy()
        self.docs['date'] = self.get_years()

        print('>> Download articles pages')
        if not os.path.exists(self.person):
            os.mkdir(self.person)
        file_dir = self.person
        websites,titles = self.docs['website'], self.docs['title']
        idx=0
        for website,title in tqdm(zip(websites,titles),total=len(websites)):
            try:
                path_file= self.get_article_page(website,file_dir,title,idx_start,idx)
                soup_article = self.get_local_soup(path_file)
                meta_data = self.get_article_metadata(soup_article)
                self.docs['authors'].append(meta_data[0].text)
                self.docs['journal'].append(meta_data[2].text)
                self.docs['abstract'].append(self.get_abstract(soup_article))
            except:
                print('problem in title', title)
                self.docs['authors'].append('')
                self.docs['journal'].append('')
                self.docs['abstract'].append('')

            idx+=1
        self.docs['field']= [self.field]*len(websites)
        print('>> Done !')







import re
from naimai.constants.regex import regex_doi, regex_words_authors, regex_words_in_brackets, regex_capital, regex_in_brackets
from naimai.constants.nlp import this_year
from naimai.constants.paths import aws_root_pdfs,doi_url,arxiv_pdfs_url
from collections import OrderedDict
import os

def lemmatize_query(nlp, query):
  text_nlp = nlp(query)
  lemmatized = []
  for token in text_nlp:
    if not token.is_stop:
        lemmatized.append(token.lemma_)
  return lemmatized

def get_nb_words(text):
    split = text.split()
    return len(split)


def multiple_replace(dictt, text):
    # Create a regular expression  from the dictionary keys
    regex = re.compile("(%s)" % "|".join(map(re.escape, dictt.keys())))

    # For each match, look-up corresponding value in dictionary
    return regex.sub(lambda mo: dictt[mo.string[mo.start():mo.end()]], text)

def reinsert_commas(with_comma,without_comma):
  split = with_comma.split(',')
  if (len(split)==3) and ',' not in without_comma:
    split=split[1]
    return without_comma.replace(split,','+split+',')
  else:
    return without_comma

def transform_field_name(field_name):
    dic_replace = {
        ' ': '%20',
        '(': '%28',
        ')': '%29'
    }
    return multiple_replace(dic_replace, field_name)

def str1_str2_from_txt(str1,str2,txt):
  pattern=str1+'(.*?)'+str2
  srch= re.search(pattern,txt, flags=re.S)
  return srch.group(1)


def correct_abbrevs_replacement(parag):
  wrds_btween_brakets = re.findall(regex_words_in_brackets,parag)
  words_repeated = [words for words in wrds_btween_brakets if len(parag.split(words))>2]
  for words in words_repeated:
    parag = re.sub(' \('+words+'\)', '',parag)
  return parag

def replace_abbreviations(pap):
    abbreviations_dict = pap.get_abbreviations_dict()
    if abbreviations_dict:
        pap.Abstract = multiple_replace(abbreviations_dict, pap.Abstract)
        pap.Abstract = correct_abbrevs_replacement(pap.Abstract)
        # pap.Conclusion = multiple_replace(abbreviations_dict, pap.Conclusion)
        # pap.Conclusion = correct_abbrevs_replacement(pap.Conclusion)

        pap.Keywords = multiple_replace(abbreviations_dict, pap.Keywords)
    return pap

def hasNumbers(inputString):
  return bool(re.search(r'\d', inputString))


def get_pattern(txt):
  pattern = r'^[\n\d\s]+\w+.*'
  patn = re.compile(pattern,flags= re.M | re.I)
  lst=re.findall(patn,txt)
  for elt in lst:
    if 'introduction' in elt.lower():
      if hasNumbers(elt):
        if '.' in elt:
          pattern = r'^\d\.[\n\s]+[a-zA-Z]+.*'
          return pattern
        else:
          pattern = r'^\d[\n\s]+[a-zA-Z]+.*'
          return pattern
  return

def starts_with_capital(text):
  regex= r'^\s*?[A-Z].*'
  result=re.findall(regex,text)
  if result:
    return True
  return False

def doi_in_text(text):
    reg = re.compile(regex_doi, re.VERBOSE + re.IGNORECASE)
    result = re.findall(reg, text)
    if result:
        try:
            return [elt for elt in result[0] if (len(elt) > 5) and ('http' not in elt)][0]
        except:
            pass
    return ''

def clean_objective_stc(stc):
  stc2=stc.replace(';', ',')
  stc3= re.sub('\(\d+\)', '', stc2)
  stc4= re.sub('\[\d+\]', 'Someone', stc3)
  stc5= re.sub('^\d+','',stc4)
  return stc5.replace('- ', '')

def clean_objectives(split):
  split_cln1 = [clean_objective_stc(stc) for stc in split]
  split_cln2= [re.sub('^\s+|\s+$','',elt) for elt in split_cln1]
  split_cln3= [re.sub('\s+',' ',elt) for elt in split_cln2 if elt]
  return split_cln3

def clean_authors_stc(stc):
  stc2=stc.replace('\n', '; ').replace("&", ',').replace('.', '. ')
  stc2 = re.sub(regex_words_authors,' ', stc2)
  return stc2

def clean_authors(split):
  split_cln1 = [clean_authors_stc(stc) for stc in split]
  split_cln2= [re.sub('^\s+|\s+$','',elt) for elt in split_cln1]
  split_cln3= [re.sub('\s+',' ',elt) for elt in split_cln2]
  return split_cln3

def check_author(author):
    split = author.split()
    if len(split)>1:
        if len(split[0])>1 and len(split[1])>1:
            return True
    return False

def fix_duplicate_authors(authors_list):
    remove_duplicates=list(set([elt.lower() for elt in authors_list]))
    capitalize_authors = [capitalize(txt) for txt in remove_duplicates]
    if authors_list:
        idx_1st_author = capitalize_authors.index(authors_list[0])
        if idx_1st_author!=0:
            new_list=[authors_list[0]]+[elt for elt in capitalize_authors if elt!=authors_list[0]]
            return new_list
        else:
            return capitalize_authors
    return

def capitalize(text):
  def cap(match):
    return (match.group().capitalize())
  pat = re.compile(r'\w+')
  return pat.sub(cap, text)

def str_in_list(str_,lst):
  found=[elt for elt in lst if str_ in elt]
  if found:
    return True
  return False

def authors_with_commas(authors_str):
    commas_split = authors_str.split(',')
    first_author = commas_split[0].split()[-1]

    if len(commas_split) > 2:
        return first_author + ' et al.'

    second = commas_split[1]

    if second:
        if len(second.split()) > 2:
            return first_author + ' et al.'

        second_author = second.split()[-1]
        return first_author + ' and ' + second_author


def authors_with_period(authors_str):
    period_split = authors_str.split('.')
    if len(period_split) == 2:
        first_author = period_split[-1]
        return re.sub(r'^\s+|\s+$', '', first_author)

    regex_authors = r'^[A-Z]\.(.*?)[A-Z]\.'
    first_author = re.findall(regex_authors, authors_str)[0]
    first_author = re.sub(r'^\s+|\s+$', '', first_author)

    split = authors_str.split(first_author)
    return authors_with_commas(split[0] + first_author + ', ' + split[1])


def authors_with_full_name(authors_str):
    if len(authors_str.split()) % 2 == 0:
        auths_split = authors_str.split()
        new_authors_list = list(
            OrderedDict.fromkeys([' '.join(auths_split[idx:idx + 2]) for idx in range(0, len(auths_split), 2)]))
        return authors_with_commas(', '.join(new_authors_list))

    return

def year_from_arxiv_fname(fname):
  yy = fname.split('.')[0][:2]
  if int('20'+yy)>this_year:
    yy = '19'+yy
  else:
    yy = '20'+yy
  return yy

def find_root_verb(sentence):
    root_token = None
    for token in sentence:
        if (token.dep_ == 'ROOT') and (token.pos_ == 'VERB'):
            root_token = token
            break
    return root_token

def filter_min_length(lst, min_len=2):
    new_list = []
    for elt in lst:
        words = re.findall(r'[a-zA-Z]+', elt)
        idx = 0
        for word in words:
            if len(word) <= min_len:
                idx += 1
        if idx < len(words):
            new_list.append(elt)
    return new_list

def get_duplicates(text,threshold):
    capital_sentences =  re.split(regex_capital,text)[1:]
    duplicate = [(elt,capital_sentences.count(elt)) for elt in capital_sentences if capital_sentences.count(elt)>threshold and len(elt.split())>1]
    duplicate.sort(key=lambda x: x[1])
    set_duplicate = set(duplicate[::-1])
    return [elt[0] for elt in set_duplicate]

def tokenId_in_txt(token,text,last_tokenId=0):
  # find token index in word level in text
  tokens = text.split()
  list_ids = [idx for idx,tok in enumerate(tokens) if tok==token and idx>=last_tokenId]
  return list_ids


def get_first_last_token_ids(sentence, text,last_tokenId=0):
    first_token = sentence.split()[0]

    first_token_id = tokenId_in_txt(first_token, text,last_tokenId=last_tokenId)
    # if len(first_token_id) > 1:
        # print(f'PROBLEM: many first token ids: {first_token_id}')
    first_token_id = first_token_id[0]

    last_token_id = first_token_id + len(sentence.split()) - 1

    return range(first_token_id, last_token_id)

def remove_between_brackets(text,nb_words=5):
  txt_in_brackets = re.findall(regex_in_brackets,text)
  txt_to_remove=[]
  if txt_in_brackets:
    for txt in txt_in_brackets:
      nb_words_txt = len(txt.split())
      if nb_words_txt>nb_words:
        rgx = re.escape('('+txt+')')
        txt_to_remove.append(rgx)

  regex_remove = '|'.join(txt_to_remove)
  return re.sub(regex_remove,'',text)

def get_ref_url(paper,doi=''):
    database = paper['database']
    if database == "mine":
        title = paper['Title']
        fname=''
        if title:
            fname = title+'.pdf'
          # fname = title.replace('/','_char_')[:100].replace("'",'_')+'.pdf'

        url = os.path.join(aws_root_pdfs, 'Geophysics', transform_field_name(fname))
        return url
    if database == "arxiv":
        url = arxiv_pdfs_url + paper['doi']
        return url
    if database =="iahr":
        url =paper['doi']
        return url
    else:
        if not doi:
            doi = paper['doi']
        url = doi_url + doi
        return url



from .direct2reported import Direct2Reported
from .clauses_processing import objective_sentence_processor
from naimai.constants.nlp import nlp_vocab
from naimai.utils.regex import authors_with_commas, authors_with_full_name
import spacy
import numpy as np

class Paper2Reported:
    def __init__(self, paper, messages, paper_name='',nlp=None):
        self.paper = paper
        self.paper_name = paper_name
        self.messages = messages
        self.paper_year = 999
        self.paper_authors = ''
        self.processed_authors=''
        self.list_reported = []
        self.reported_objective = ''
        self.objective_queue = []
        if nlp:
            self.nlp = nlp
        else:
            self.nlp = spacy.load(nlp_vocab)

    def get_paper_year(self):
        paper_year = self.paper['year']
        if paper_year != 999:
            self.paper_year = paper_year
        else:
            self.paper_year = ''

    def get_paper_authors(self):
        if self.paper['Authors']:
            self.paper_authors = self.paper['Authors']
        else:
            self.paper_authors = 'Some authors'

    def process_authors(self):
        str_year = ' (' + str(self.paper_year) + ')'
        authors = self.paper_authors
        result = ''
        if len(authors.split()) == 2:
            result = authors.split()[1] + str_year
        else:
            try:
                if ',' in authors:
                    result = authors_with_commas(authors) + str_year
                else:
                    authors_formula = authors_with_full_name(authors)
                    if authors_formula:
                        if "." in authors_formula:
                            result = authors_formula + str_year
                        else:
                            result = authors_with_full_name(authors) + str_year
                    else:
                        result = 'Some authors ' + str_year
            except:
                print('problem authors with paper of doi {} - dbase {} '.format(self.paper['doi'],self.paper['database']))
        self.processed_authors= result

    # def choose_objective(self):
    #     len_objs = len(self.messages)
    #     idx_obj = 0
    #     if len_objs > 1:
    #         len_obj_elts = [len(el.split()) for el in self.messages]
    #         if self.complexity:  # long objectives
    #             idx_obj = len_obj_elts.index(max(len_obj_elts))
    #         else:
    #             idx_obj = len_obj_elts.index(min(len_obj_elts))
    #     chosen = self.messages[idx_obj]
    #     self.objective_queue = [obj for obj in self.messages if obj != chosen]
    #     return chosen

    # def gather_authors_objectives(self):
    #     if self.messages:
    #         self.get_paper_authors()
    #         authors = self.process_authors()
    #         self.get_paper_year()
    #         paper_obj = self.choose_objective()
    #         return (authors, paper_obj)
    #     return
    #
    # def change_objective(self):
    #     chosen = self.objective_queue[0]
    #     self.objective_queue.remove(chosen)
    #     return chosen

    def generate(self):
        writer = None
        self.get_paper_year()
        self.get_paper_authors()
        self.process_authors()


        for obj in self.messages:
            try :
                obj_processor = objective_sentence_processor(sentence=obj, nlp=self.nlp)
                obj_processor.process()
                obj_transformed = obj_processor.final_sentence
            except:
                obj_transformed = obj
            try:
                writer = Direct2Reported(authors=self.processed_authors, sentence=obj_transformed, nlp=self.nlp)
                writer.generate()
                if writer.reported:
                    self.list_reported.append(writer.reported)
                else:
                    writer = Direct2Reported(authors=self.processed_authors, sentence=obj, nlp=self.nlp)
                    writer.generate()
                    self.list_reported.append(writer.reported)
            except:
                with open('objectives_pbs.txt', 'a') as f:
                    f.write('problem with objective : {} // paper : {} - dbase : {} \n\n'.format(obj,self.paper_name,self.paper['database']))

    def choose_obj_in_mean_lengths_range(self,mean_objs):
        mean_lens = 22
        if len(mean_objs) == 1:
            return mean_objs[0]
        else:
            # take the closest len to the mean 22
            closest_to_mean = min(mean_objs, key=lambda x: abs(len(x.split()) - mean_lens))
            return closest_to_mean

    def choose_obj_beyond_mean_lengths_range(self, references):
        mean_lens_radius = np.arange(20, 26)
        closest_to_min_radius = min(references, key=lambda x: abs(len(x.split()) - mean_lens_radius[0]))
        return closest_to_min_radius

    def choose_objective(self):
        mean_lens_radius = np.arange(20, 26)
        if len(self.list_reported) == 1:  # if only one ref, take it
            self.reported_objective= self.list_reported[0]
        if len(self.list_reported) > 1:  # keep lengths in mean lengths radius 20-25
            mean_objs = [elt for elt in self.list_reported if len(elt.split()) in mean_lens_radius]
            if mean_objs:  # we have objs in the lengths radius:
                self.reported_objective = self.choose_obj_in_mean_lengths_range(mean_objs)
            else:  # so len(references)>1 and lengths of all refs are either < 20 (so we take the one with max length)
                # or > 26 (so we take the one with min length)
                self.reported_objective=  self.choose_obj_beyond_mean_lengths_range(self.list_reported)



from naimai.constants.paths import path_produced
from naimai.utils.general import clean_lst
from .sqlite_manager import SQLiteManager
import os
import re
import spacy
from naimai.constants.nlp import nlp_vocab
from naimai.constants.regex import regex_and_operators,regex_exact_match
from naimai.constants.models import threshold_tf_similarity
from naimai.models.papers_classification.tfidf import tfidf_model

class BaseQuerier:
  def __init__(self,field='', encoder=None, field_index=None,nlp=None,is_custom=True):
    self.encoder = encoder
    self.default_top_n = 200
    self.field = field
    self.field_index = field_index
    self.is_custom = is_custom

    if field:
      path_sqlite = os.path.join(path_produced, self.field, 'all_papers_sqlite')
      self.sql_manager = SQLiteManager(path_sqlite)
    self.search_operators = {'single': 0, "multiple": 1, "match":2}
    self.nlp = nlp
    self.load_nlp()

  def load_nlp(self):
      if not self.nlp:
          print('>> Loading nlp..')
          self.nlp = spacy.load(nlp_vocab)

  def remove_duplicated_fnames(self,list_fnames):
      '''
      remove duplicated fnames in list of results to keep only one paper per result
      :param list_fnames:
      :return:
      '''
      pattern = '_objectives|_methods|_results'
      only_fnames = [re.sub(pattern, '', elt) for elt in list_fnames]
      filtered_fnames = clean_lst(only_fnames)

      idxs_to_keep = []
      for fname in filtered_fnames:
        idxs_to_keep.append([idx for idx, elt in enumerate(list_fnames) if fname in elt][0])

      non_duplicated_fnames = [list_fnames[idx] for idx in idxs_to_keep]
      return non_duplicated_fnames


  # def get_corresponding_papers(self, wanted_papers_fnames: list, root_papers_fnames: list) -> list:
  #   '''
  #   XXX
  #   :param wanted_papers_fnames:
  #   :param root_papers_fnames:
  #   :return:
  #   '''
  #   root_fnames = [elt.replace('_objectives', '') for elt in root_papers_fnames]
  #
  #   results_papers_fnames = []
  #   for elt in root_fnames:
  #     for idx, pap in enumerate(wanted_papers_fnames):
  #       if elt in pap:
  #         results_papers_fnames.append(wanted_papers_fnames[idx])
  #
  #   return results_papers_fnames

  def sort_using_citations(self, papers: dict,top_n=5) -> list:
    '''
    Rank papers papers using numCitedBy parameter.
    :param papers:
    :return:
    '''
    nb_papers_ranked_CitedBy = len(papers)

    for fname in papers:
      if 'numCitedBy' not in papers[fname]:
        papers[fname]['numCitedBy'] = .5
    keys = list(papers.keys())
    papers_to_rank = [papers[fname] for fname in papers]
    papers_to_rank.sort(key=lambda x: float(x['numCitedBy']), reverse=True)
    fnames_ranked = [elt['fname'] for elt in papers_to_rank]
    rest_of_papers_fnames = [fname for fname in keys[nb_papers_ranked_CitedBy:]]
    papers_ranked = fnames_ranked + rest_of_papers_fnames
    return papers_ranked[:top_n]

  def sort_using_dates(self, papers, top_n) -> list:
    '''
    rank papers using their publications date
    :param papers:
    :param top_n:
    :return:
    '''

    nb_papers = len(papers)

    keys = list(papers.keys())
    papers_to_rank = [papers[fname] for fname in papers]
    papers_to_rank.sort(key=lambda x: float(x['year']), reverse=True)
    fnames_ranked = [elt['fname'] for elt in papers_to_rank]
    rest_of_papers_fnames = [fname for fname in keys[nb_papers:]]
    papers_ranked = fnames_ranked + rest_of_papers_fnames
    return papers_ranked[:top_n]

  def get_query_type(self, query: str) -> int:
    '''
    determine the query type : Simple operator (simple keywords), AND operator or exact match. OR operator is removed for the moment since it's rarely used.
    '''
    AND_operator = re.findall(regex_and_operators, query)
    exact_match = re.findall(regex_exact_match, query)

    if AND_operator:
      return self.search_operators['multiple']
    if exact_match:
      return self.search_operators['match']
    return self.search_operators['single']

  def get_papers_with_AND_operator(self,query,year_from=0,year_to=3000,top_n=200):
    '''
    overwrited method in custom_querier and with_keywords class
    '''
    return (None,None)

  def get_papers_for_tfidf_semantics(self,query,year_from=0,year_to=3000,top_n=200):
    '''
    overwrited method in custom_querier and with_keywords class
    '''
    return (None,None)

  def get_papers_with_exact_match(self,query,year_from=0,year_to=3000,top_n=200):
    '''
    overwrited method in custom_querier and with_keywords class
    '''
    return (None, None)

  def get_relevant_papers(self, query: str, query_type: int, year_from=0, year_to=3000, top_n=5) -> tuple:
    '''
    Get all similar papers & their fnames based on the query and query type. Here, we return tuple.
    Start with semantic search. If an operator is used, get the first 200 papers > apply operator
    :param query:
    :return:
    '''

    selected_papers, selected_papers_fnames= [],[]
    if query_type == self.search_operators['multiple']:  # AND operator
      selected_papers, selected_papers_fnames = self.get_papers_with_AND_operator(query,year_from=year_from,year_to=year_to,top_n=top_n)

    # elif query_type == 1:  # OR operator
    #   selected_papers, selected_papers_fnames = self.get_papers_with_OR_operator(query,year_from=year_from,year_to=year_to,top_n=top_n)

    elif query_type == self.search_operators['match']:  # exact match
      selected_papers, selected_papers_fnames = self.get_papers_with_exact_match(query,year_from=year_from,year_to=year_to,top_n=top_n)

    elif query_type == self.search_operators['single']:  # simple operator
      selected_papers, selected_papers_fnames = self.get_papers_for_tfidf_semantics(query,year_from=year_from,year_to=year_to,top_n=top_n)

    return selected_papers, selected_papers_fnames

  def sort_results(self,papers: dict,query: str,method='pertinence',top_n=5) -> list:
    '''
    Sort by 'pertinence', 'citations' or 'date'
    :param papers:
    :param query:
    :param method:
    :param top_n:
    :return:
    '''
    if self.is_custom:
      sorted_papers_fnames = [fname for fname in papers]
    else:
      sorted_papers_fnames = self.sort_using_tf_model(query, papers, top_n)

    if method == 'pertinence':
        return sorted_papers_fnames
    new_papers = {name: papers[name] for name in sorted_papers_fnames}

    if method =='citations':
        sorted_papers_fnames = self.sort_using_citations(new_papers,top_n)
        return sorted_papers_fnames
    elif method =='date':
        sorted_papers_fnames = self.sort_using_dates(new_papers, top_n)
        return sorted_papers_fnames
    return sorted_papers_fnames

  def sort_using_tf_model(self,query: str,papers: dict,top_n: int) -> list:
      tf = tfidf_model(query=query, papers=papers)
      tf.vectorizer.min_df = .05
      tf_ranked_papers_fnames, scores = tf.get_similar_fnames(top_n=top_n)

      if scores[0] < threshold_tf_similarity:
          print('>> WARNING : These results may not be relevant!')
      return tf_ranked_papers_fnames


  def find_papers(self, query: str, top_n=5, year_from=0, year_to=3000, verbose=False,sort_by='pertinence') -> list:
      '''
      Sort by 'pertinence', 'citations' or 'date'
      1. Get query type : simple operator, AND operator or exact match.
      2. Get 200 relevant papers and their fnames following the operator type
      3. Classify using a sorting method
      '''

      # 1. Get query type : simple operator, AND operator or exact match.
      query_type = self.get_query_type(query)
      if verbose:
          operator = [elt for elt in self.search_operators if self.search_operators[elt]==query_type][0]
          print('Operator: ', operator)

      # 2. Get all relevant papers and their fnames
      selected_papers, selected_papers_fnames = self.get_relevant_papers(query, query_type, year_from=year_from, year_to=year_to, top_n=self.default_top_n)
      if len(selected_papers)>self.default_top_n:
          print(f'More than {self.default_top_n} papers!')

      # 3. Classify using a sorting method
      if selected_papers:
          sorted_fnames = self.sort_results(selected_papers,query,sort_by,top_n)

          if verbose:
              print(' ')
          return sorted_fnames
      return []

  def review(self,query: str, top_n=100, sort_by='pertinence'):
    '''
    Generate a review
    :param query:
    :param top_n:
    :param sort_by:
    :return:
    '''

    papers = self.find_papers(query=query,top_n=top_n, sort_by=sort_by, verbose=False)

    # Get text
    reported_list = ['- ' + pap['reported'] for pap in papers if 'reported' in pap]
    review_text = '\n'.join(reported_list)

    # Get references
    references_list =[]
    for pap in papers:
      if 'reported' in pap and pap['reported']:
        paper_refs = '- ' + self.get_reference(pap)
        references_list.append(paper_refs)
    references_text =  '\n'.join(references_list)
    return review_text,references_text

  def get_reference(self, paper_objectives: dict) -> str:
    '''get paper references in format : Authors + year + Title + Journal, using root paper'''

    authors, year, title, journal = '', '', '', ''
    if 'allauthors' in paper_objectives:
      authors = paper_objectives['allauthors']

    if 'year' in paper_objectives:
      year = paper_objectives['year']

    if 'title' in paper_objectives:
      title = paper_objectives['title']

    if 'journal' in paper_objectives:
      journal = paper_objectives['journal']

    reference = authors + ', ' + str(year) + ', ' + title + ', ' + journal + '.'
    return reference


import regex
from collections import defaultdict, Counter

"""
A Python 3 refactoring of Vincent Van Asch's Python 2 code at
http://www.cnts.ua.ac.be/~vincent/scripts/abbreviations.py
Based on
A Simple Algorithm for Identifying Abbreviations Definitions in Biomedical Text
A. Schwartz and M. Hearst
Biocomputing, 2003, pp 451-462.
"""

class Candidate(str):
    def __init__(self, value):
        super().__init__()
        self.start = 0
        self.stop = 0

    def set_position(self, start, stop):
        self.start = start
        self.stop = stop


def yield_lines_from_file(file_path):
    with open(file_path, 'rb') as f:
        for line in f:
            try:
                line = line.decode('utf-8')
            except UnicodeDecodeError:
                line = line.decode('latin-1').encode('utf-8').decode('utf-8')
            line = line.strip()
            yield line


def yield_lines_from_doc(doc_text):
    for line in doc_text.split("\n"):
        yield line.strip()


def best_candidates(sentence):
    """
    :param sentence: line read from input file
    :return: a Candidate iterator
    """

    if '(' in sentence:
        # Check some things first
        if sentence.count('(') != sentence.count(')'):
            raise ValueError("Unbalanced parentheses: {}".format(sentence))

        if sentence.find('(') > sentence.find(')'):
            raise ValueError("First parentheses is right: {}".format(sentence))

        close_index = -1
        while 1:
            # Look for open parenthesis. Need leading whitespace to avoid matching mathematical and chemical formulae
            open_index = sentence.find(' (', close_index + 1)

            if open_index == -1: break

            # Advance beyond whitespace
            open_index += 1

            # Look for closing parentheses
            close_index = open_index + 1
            open_count = 1
            skip = False
            while open_count:
                try:
                    char = sentence[close_index]
                except IndexError:
                    # We found an opening bracket but no associated closing bracket
                    # Skip the opening bracket
                    skip = True
                    break
                if char == '(':
                    open_count += 1
                elif char in [')', ';', ':']:
                    open_count -= 1
                close_index += 1

            if skip:
                close_index = open_index + 1
                continue

            # Output if conditions are met
            start = open_index + 1
            stop = close_index - 1
            candidate = sentence[start:stop]

            # Take into account whitespace that should be removed
            start = start + len(candidate) - len(candidate.lstrip())
            stop = stop - len(candidate) + len(candidate.rstrip())
            candidate = sentence[start:stop]

            if conditions(candidate):
                new_candidate = Candidate(candidate)
                new_candidate.set_position(start, stop)
                yield new_candidate


def conditions(candidate):
    """
    Based on Schwartz&Hearst
    2 <= len(str) <= 10
    len(tokens) <= 2
    re.search(r'\p{L}', str)
    str[0].isalnum()
    and extra:
    if it matches (\p{L}\.?\s?){2,}
    it is a good candidate.
    :param candidate: candidate abbreviation
    :return: True if this is a good candidate
    """
    viable = True
    if regex.match(r'(\p{L}\.?\s?){2,}', candidate.lstrip()):
        viable = True
    if len(candidate) < 2 or len(candidate) > 10:
        viable = False
    if len(candidate.split()) > 2:
        viable = False
    if not regex.search(r'\p{L}', candidate):
        viable = False
    if not candidate[0].isalnum():
        viable = False

    return viable


def get_definition(candidate, sentence):
    """
    Takes a candidate and a sentence and returns the definition candidate.
    The definition candidate is the set of tokens (in front of the candidate)
    that starts with a token starting with the first character of the candidate
    :param candidate: candidate abbreviation
    :param sentence: current sentence (single line from input file)
    :return: candidate definition for this abbreviation
    """
    # Take the tokens in front of the candidate
    tokens = regex.split(r'[\s\-]+', sentence[:candidate.start - 2].lower())
    # the char that we are looking for
    key = candidate[0].lower()

    # Count the number of tokens that start with the same character as the candidate
    first_chars = [t[0] for t in filter(None, tokens)]

    definition_freq = first_chars.count(key)
    candidate_freq = candidate.lower().count(key)

    # Look for the list of tokens in front of candidate that
    # have a sufficient number of tokens starting with key
    if candidate_freq <= definition_freq:
        # we should at least have a good number of starts
        count = 0
        start = 0
        start_index = len(first_chars) - 1
        while count < candidate_freq:
            if abs(start) > len(first_chars):
                raise ValueError("candidate {} not found".format(candidate))
            start -= 1
            # Look up key in the definition
            try:
                start_index = first_chars.index(key, len(first_chars) + start)
            except ValueError:
                pass

            # Count the number of keys in definition
            count = first_chars[start_index:].count(key)

        # We found enough keys in the definition so return the definition as a definition candidate
        start = len(' '.join(tokens[:start_index]))
        stop = candidate.start - 1
        candidate = sentence[start:stop]

        # Remove whitespace
        start = start + len(candidate) - len(candidate.lstrip())
        stop = stop - len(candidate) + len(candidate.rstrip())
        candidate = sentence[start:stop]

        new_candidate = Candidate(candidate)
        new_candidate.set_position(start, stop)
        return new_candidate

    else:
        raise ValueError('There are less keys in the tokens in front of candidate than there are in the candidate')


def select_definition(definition, abbrev):
    """
    Takes a definition candidate and an abbreviation candidate
    and returns True if the chars in the abbreviation occur in the definition
    Based on
    A simple algorithm for identifying abbreviation definitions in biomedical texts, Schwartz & Hearst
    :param definition: candidate definition
    :param abbrev: candidate abbreviation
    :return:
    """

    if len(definition) < len(abbrev):
        raise ValueError('Abbreviation is longer than definition')

    if abbrev in definition.split():
        raise ValueError('Abbreviation is full word of definition')

    s_index = -1
    l_index = -1

    while 1:
        try:
            long_char = definition[l_index].lower()
        except IndexError:
            raise

        short_char = abbrev[s_index].lower()

        if not short_char.isalnum():
            s_index -= 1

        if s_index == -1 * len(abbrev):
            if short_char == long_char:
                if l_index == -1 * len(definition) or not definition[l_index - 1].isalnum():
                    break
                else:
                    l_index -= 1
            else:
                l_index -= 1
                if l_index == -1 * (len(definition) + 1):
                    raise ValueError("definition {} was not found in {}".format(abbrev, definition))

        else:
            if short_char == long_char:
                s_index -= 1
                l_index -= 1
            else:
                l_index -= 1

    new_candidate = Candidate(definition[l_index:len(definition)])
    new_candidate.set_position(definition.start, definition.stop)
    definition = new_candidate

    tokens = len(definition.split())
    length = len(abbrev)

    if tokens > min([length + 5, length * 2]):
        raise ValueError("did not meet min(|A|+5, |A|*2) constraint")

    # Do not return definitions that contain unbalanced parentheses
    if definition.count('(') != definition.count(')'):
        raise ValueError("Unbalanced parentheses not allowed in a definition")

    return definition


def extract_abbreviation_definition_pairs(file_path=None,
                                          doc_text=None,
                                          most_common_definition=False,
                                          first_definition=False):
    abbrev_map = dict()
    list_abbrev_map = defaultdict(list)
    counter_abbrev_map = dict()
    omit = 0
    written = 0
    if file_path:
        sentence_iterator = enumerate(yield_lines_from_file(file_path))
    elif doc_text:
        sentence_iterator = enumerate(yield_lines_from_doc(doc_text))
    else:
        return abbrev_map

    collect_definitions = False
    if most_common_definition or first_definition:
        collect_definitions = True

    for i, sentence in sentence_iterator:
        # Remove any quotes around potential candidate terms
        clean_sentence = regex.sub(r'([(])[\'"\p{Pi}]|[\'"\p{Pf}]([);:])', r'\1\2', sentence)
        try:
            for candidate in best_candidates(clean_sentence):
                try:
                    definition = get_definition(candidate, clean_sentence)
                except (ValueError, IndexError) as e:
                    omit += 1
                else:
                    try:
                        definition = select_definition(definition, candidate)
                    except (ValueError, IndexError) as e:
                        omit += 1
                    else:
                        # Either append the current definition to the list of previous definitions ...
                        if collect_definitions:
                            list_abbrev_map[candidate].append(definition)
                        else:
                            # Or update the abbreviations map with the current definition
                            abbrev_map[candidate] = definition
                        written += 1
        except (ValueError, IndexError) as e:
          pass

    # Return most common definition for each term
    if collect_definitions:
        if most_common_definition:
            # Return the most common definition for each term
            for k,v in list_abbrev_map.items():
                counter_abbrev_map[k] = Counter(v).most_common(1)[0][0]
        else:
            # Return the first definition for each term
            for k, v in list_abbrev_map.items():
                counter_abbrev_map[k] = v[0]
        return counter_abbrev_map

    # Or return the last encountered definition for each term
    return abbrev_map


import requests
from bs4 import BeautifulSoup
from tqdm.notebook import tqdm
from naimai.constants.regex import regex_not_converted2
import re
import time
import random

class EGU_Crawler_Page:
    def __init__(self,t_min=2,t_max=4):
        self.docs = {'title': [], 'authors': [], 'date': [], "abstract": [], "doi": [],
                     "field": []}
        self.t_min = t_min
        self.t_max = t_max

    def get_soup(self, path):
        header = {}
        header[
            'User-Agent'] = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'
        soup = BeautifulSoup(requests.get(path, headers=header, timeout=15).content, 'html.parser')
        slp=random.randint(self.t_min,self.t_max)
        time.sleep(slp)
        return soup

    def get_abstract(self,soup):
        abstract_div = soup.find_all(name='div', attrs={'class': 'abstract'})
        try:
            abstract = abstract_div[1].text.replace('\nAbstract.', '')
            abstract = re.sub(regex_not_converted2,' ',abstract).strip()
        except:
            abstract = ''
        return abstract

    def get_title(self,soup):
        title_div = soup.find_all(name='h1')
        title = title_div[1].text.replace('\n','').replace('\r',' ').strip()
        return title

    def get_authors(self,soup):
        authors_div = soup.find(name='strong', attrs={'class': 'hide-on-mobile'})
        authors_div2 = authors_div.find_all(name='nobr')
        authors = [re.sub('\d,?', '', elt.text).replace(',', '') for elt in authors_div2]
        return authors

    def get_date(self,doi):
        return doi.split('-')[-1]

    def get_doc(self,doi):
        doc = {}
        soup = self.get_soup(doi)
        doc['abstract'] = self.get_abstract(soup)
        doc['title'] = self.get_title(soup)
        doc['authors'] = self.get_authors(soup)
        doc['date'] = self.get_date(doi)
        doc['field'] = 'Environmental Science'
        return doc


class EGU_Crawler:
    def __init__(self,t_min=2,t_max=4):
        self.docs = {'title': [], 'authors': [], 'date': [], "abstract": [], "doi": [],
                     "field": []}
        self.soups={}
        self.t_min = t_min
        self.t_max = t_max
        self.page_crawler = EGU_Crawler_Page(t_min=t_min,t_max=t_max)

    def get_soup(self, path,data):
        page = requests.post(path, data=data)
        soup = BeautifulSoup(page.text, 'html.parser')
        slp=random.randint(self.t_min,self.t_max)
        time.sleep(slp)
        return soup

    def get_soups(self,p1,p2,type_):
        data = {
            'rangeMin': 0,
            'rangeMax': 7,
            'manuscriptTypes[]': type_,
        }
        path = 'https://editor.copernicus.org/ms_types.php?journalId=10'
        for page in tqdm(range(p1,p2+1)):
            data['page']=page-1
            self.soups[page]= self.get_soup(path,data)

    def get_title(self,div):
        title_div = div.find(name='a', attrs={'class': 'article-title'})
        titles = title_div.text
        return titles

    def get_date(self,div):
        date_div = div.find(name='div', attrs={'class': 'published-date'})
        date = date_div.text.split()[-1]
        return date

    def get_authors(self,div):
        authors_div = div.find(name='div', attrs={'class': 'authors'})
        authors = authors_div.text
        return authors

    def get_doi(self,div):
        doi_div = div.find(name='div', attrs={'class': 'citation'})
        doi = doi_div.find(name='span').text.replace(',', '').strip()
        return doi

    def get_abstract(self,div):
        abstract_div = div.find(name='div', attrs={'class': 'content'})
        if abstract_div:
            return abstract_div.text.replace('\n', '').strip()
        return ''

    def get_divs(self,soup):
        divs = soup.find_all(name='div', attrs={'class': 'grid-container paperlist-object in-range paperList-final'})
        return divs



    def get_docs(self,p1=1, p2=10,type_=22):
        print('>> Getting soups..')
        self.get_soups(p1=p1,p2=p2,type_=type_)

        print('>> Getting articles..')
        for page in tqdm(self.soups):
            soup = self.soups[page]
            articles_divs=self.get_divs(soup)

            for div in articles_divs:
                abstract = self.get_abstract(div)
                doi = self.get_doi(div)
                if abstract:
                    title = self.get_title(div)
                    date = self.get_date(div)
                    authors = self.get_authors(div)
                    fields = 'Environmental Science'

                    self.docs['title'].append(title)
                    self.docs['date'].append(date)
                    self.docs['abstract'].append(abstract)
                    self.docs['doi'].append(doi)
                    self.docs['authors'].append(authors)
                    self.docs['field'].append(fields)
                else:
                    paper = self.page_crawler.get_doc(doi)
                    if paper['abstract']:
                        self.docs['title'].append(paper['title'])
                        self.docs['date'].append(paper['date'])
                        self.docs['abstract'].append(paper['abstract'])
                        self.docs['doi'].append(doi)
                        self.docs['authors'].append(paper['authors'])
                        self.docs['field'].append(paper['field'])


from naimai.constants.regex import regex_and_operators,regex_exact_match
from naimai.models.papers_classification.tfidf import tfidf_model
from naimai.utils.regex import lemmatize_query
from .base import BaseQuerier
import re

class CustomQuerier(BaseQuerier):
  def __init__(self,produced_papers):
    super().__init__(is_custom=True)
    self.produced_papers = produced_papers

  def keywords_in_paper(self,keywords: list,operator: int,paper: dict) -> bool:
    '''
    check if keywords are in papers following an operator
    '''   
    fname = list(paper.keys())[0]

    if '_objectives' in fname:
      info = '. '.join(self.produced_papers[fname]['messages']) + ' '+ self.produced_papers[fname]['title']
    else:
      info = '. '.join(self.produced_papers[fname]['messages'])

    if operator==self.search_operators['multiple']: #AND operator
      pattern = ''.join([f'(?:.*{kw})'for kw in keywords])
      if re.findall(pattern,info,flags=re.I):
        return True

    elif operator==self.search_operators['match']: # exact match
      for query in keywords:
        if re.findall(query,info):
          return True

    return False 

  def get_papers_with_exact_match(self,query: str,year_from=0,year_to=3000,top_n=200) -> tuple:
    '''
    find papers for a query with exact match
    '''
    keywords = re.findall(regex_exact_match, query)
    selected_papers_fnames = []
    selected_papers = []

    for fname in self.produced_papers:
      paper = {fname : self.produced_papers[fname]}
      if self.keywords_in_paper(keywords=keywords,operator=2,paper=paper):
        selected_papers_fnames.append(fname)
        selected_papers.append(paper)
    return selected_papers, selected_papers_fnames


  def get_papers_with_AND_operator(self,query: str,year_from=0,year_to=3000,top_n=200) -> tuple:
    '''
    find papers for a query with and operator
    '''

    keywords = [elt.strip() for elt in re.split(regex_and_operators,query)]
    selected_papers_fnames = []
    selected_papers = []

    for fname in self.produced_papers:
      paper = {fname : self.produced_papers[fname]}
      if self.keywords_in_paper(keywords=keywords,operator=0,paper=paper):
        selected_papers_fnames.append(fname)
        selected_papers.append(paper)
    return selected_papers, selected_papers_fnames


  def get_papers_for_tfidf_semantics(self,query: str, year_from=0, year_to=3000, top_n=30) -> tuple:
    '''
    find papers using tf idf model
    '''

    lemmatized_query = ' '.join(lemmatize_query(self.nlp, query)).strip()
    tf = tfidf_model(query=lemmatized_query, papers=self.produced_papers)
    selected_papers_fnames,_ = tf.get_similar_fnames(top_n=top_n)
    selected_papers = [self.produced_papers[fn] for fn in selected_papers_fnames]
    return selected_papers, selected_papers_fnames


from collections import Counter
from transformers import Trainer
from tqdm.notebook import tqdm
import re
import numpy as np
import pandas as pd
from naimai.utils.transformers import score_feedback_comp
from naimai.constants.models import ner_labels, output_labels
from naimai.constants.paths import path_detailed_data_total
from torch import cuda

class Predictions_preparer:
    '''
    transform prediction to pandas to use new metric for scoring
    '''

    def __init__(self, predictions, tokenizer, datasets):
        '''
        predictions = tensor([[1,2,3], [6,5,1]]) & classifier
        :param predictions: output of model (logits.argmax(-1))
        :param tokenizer: tokenizer of model
        :param datasets: Dataset objects (datasets['test'])
        '''
        self.predictions = predictions
        self.predictions_df = None
        self.tokenizer = tokenizer
        self.datasets = datasets
        self.ids2labels = {k: v for k, v in enumerate(output_labels)}

    def remove_tokenization_effect(self, prediction, tokens):
        '''
        when tokenized, sub words are labelled too and we end up with more word ids than the input. We remove this tokenization effect here and return the prediction filtered
        :param prediction:
        :param tokens:
        :return:
        '''
        if cuda.is_available():
            encoding = self.tokenizer(tokens, truncation=True, is_split_into_words=True, return_tensors='pt').to('cuda')
        else:
            encoding = self.tokenizer(tokens, truncation=True, is_split_into_words=True, return_tensors='pt')

        predictions_list = prediction.tolist()
        predictions_filtered = []
        words_ids = encoding.word_ids()
        previous_word_id = None
        for idx, word_id in enumerate(words_ids):
            if word_id is None:
                pass
            elif word_id != previous_word_id:
                try:
                    predictions_filtered.append(predictions_list[idx])
                    previous_word_id = word_id
                except:
                    # print(f'idx {i}, widx = {word_idx}, label {label}, len word ids= {len(word_ids)}')
                    predictions_filtered.append(0)
                    previous_word_id = word_id

        return predictions_filtered

    def prediction_filtered2named_labels(self, prediction_filtered):
        '''
        convert prediction ids to named labels
        :param prediction_filtered:
        :return:
        '''
        return [self.ids2labels[elt] for elt in prediction_filtered]

    def clean_named_labels(self, named_labels):
        '''
        remove the initial B- & I- from labels
        :param named_labels: list named labels
        :return:
        '''
        result = [re.sub('[BI]\-', '', elt) for elt in named_labels]
        result = [elt.replace('O', 'other') for elt in result]
        return result

    def get_wids_classes(self, clean_labels):
        '''
        get word ids (or predictionstring) and classes from clean labels
        :param clean_labels:
        :return:
        '''
        cnt = Counter(clean_labels)
        list_prediction_string = []
        classes = []
        for key in cnt:
            prediction_string = [str(idx + 1) for idx, elt in enumerate(clean_labels) if elt == key]
            list_prediction_string.append(' '.join(prediction_string))
            classes.append(key)
        return {'class': classes, 'predictionstring': list_prediction_string}

    def to_df(self, doi, wids_classes):
        '''
         clf.trainer.eval_dataset['doi']
        :param doi:
        :param wids_classes:
        :return:
        '''
        wids_classes['doi'] = [doi] * len(wids_classes['class'])
        return pd.DataFrame(wids_classes)

    def prepare_one_prediction(self,doi,prediction,tokens):
        '''
        transform one prediction into dataframe with columns : class, predictionstring & doi
        :param doi: doi in str
        :param prediction: prediction output
        :param tokens: tokens of predicted text
        :return:
        '''
        prediction_filtered = self.remove_tokenization_effect(prediction, tokens)
        prediction_named = self.prediction_filtered2named_labels(prediction_filtered)
        clean_prediciton_named = self.clean_named_labels(prediction_named)
        wids_classes = self.get_wids_classes(clean_prediciton_named)
        df = self.to_df(doi, wids_classes)
        return df

    def prepare(self, show_progress=False):
        list_df = []
        if show_progress:
            range_ = tqdm(zip(self.predictions, self.datasets), total=len(self.predictions))
        else:
            range_ = zip(self.predictions, self.datasets)
        for prediction, dataset in range_:
            doi, tokens = dataset['doi'], dataset['tokens']
            df = self.prepare_one_prediction(doi=doi,prediction=prediction,tokens=tokens)
            list_df.append(df)
        self.predictions_df = pd.concat(list_df)

class BOMR_Trainer(Trainer):

    def __init__(self ,*args ,**kwargs):
        super().__init__(*args ,**kwargs)

    def evaluation_loop(
            self,
            dataloader,
            description,
            prediction_loss_only = None,
            ignore_keys = None,
            metric_key_prefix = "eval",
    ):
        eval_output = super().evaluation_loop(
            dataloader,
            description,
            prediction_loss_only,
            ignore_keys,
            metric_key_prefix
        )

        eval_dataset = self.eval_dataset
        # new_metrics = {}
        is_in_eval = metric_key_prefix == "eval"
        if is_in_eval:
            predictions = eval_output.predictions.argmax(-1)
            preparer = Predictions_preparer(predictions=predictions,
                                            tokenizer=self.tokenizer,
                                            datasets=eval_dataset)

            preparer.prepare()
            eval_pred_df = preparer.predictions_df
            ground_truth_df = pd.read_csv(path_detailed_data_total).dropna()
            eval_gt_df = ground_truth_df[ground_truth_df["doi"].isin(eval_dataset["doi"])].reset_index(drop=True).copy()

            list_class_f1scores = []
            for class_ in ner_labels:
                gt_df = eval_gt_df.loc[eval_gt_df['class'] == class_].copy()
                pred_df = eval_pred_df.loc[eval_pred_df['class'] == class_].copy()
                f1_score = score_feedback_comp(pred_df, gt_df)
                eval_output.metrics[f"eval_F1-{class_}"] =f1_score
                # new_metrics[f"eval_F1-{class_}"] =f1_score
                list_class_f1scores.append(f1_score)

            eval_output.metrics["eval_F1-avg"] = np.mean(list_class_f1scores)
            # new_metrics["eval_F1-avg"] = np.mean(list_class_f1scores)
        return eval_output
        # return EvalLoopOutput(predictions = eval_output.predictions,
        #                       label_ids = eval_output.label_ids,
        #                       metrics=new_metrics,
        #                       num_samples=eval_output.num_samples)


import re
import os
from tqdm.notebook import tqdm
from bs4 import BeautifulSoup
import random
import requests
import time

years_dir = "issue/browse-by-year"
root_dir ="https://iwaponline.com"

class IWA_Crawler:
    def __init__(self, journal_path,t_min=3,t_max=5):
        self.journal_path = journal_path
        self.docs = {'title': [], 'abstract': [], 'authors': [], 'date': [], 'field': [], "keywords":[],"doi":[],"numCitedBy":[]}
        self.t_min=t_min
        self.t_max=t_max

    def get_soup(self, path):
        header = {}
        header[
            'User-Agent'] = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'
        soup = BeautifulSoup(requests.get(path, headers=header, timeout=30).content, 'html.parser')
        slp=random.randint(self.t_min,self.t_max)
        time.sleep(slp)
        return soup

    def get_years(self):
        path = os.path.join(self.journal_path,years_dir)
        journal_years_soup = self.get_soup(path)
        years_div = journal_years_soup.find_all(name='ol')
        years = [elt.text.replace('\n','').strip() for elt in years_div[0].find_all('li')]
        print(f'From {years[0]} - To {years[-1]} - Total years : {len(years)} ')
        return years

    def get_year_volumes(self,year):
        path = os.path.join(self.journal_path,years_dir,str(year))
        year_soup = self.get_soup(path)
        volumes_divs = year_soup.find_all(name='ol')[0].find_all('li')
        volumes_dirs = [elt.find(name='a')['href'] for elt in volumes_divs]
        return volumes_dirs

    def get_articles_soups_in_volume(self,volume_dir):
        path = root_dir + volume_dir
        soup_issue_articles = self.get_soup(path)
        titles_divs = soup_issue_articles.find_all(name='div', attrs={'class': 'al-article-item-wrap al-normal'})
        articles_href = [elt.find(name='h5').find(name='a')['href'] for elt in titles_divs]
        articles_paths = [root_dir + elt for elt in articles_href]
        articles_soups = [self.get_soup(p) for p in articles_paths]
        return articles_soups

    def get_main_content_article(self,article_soup):
        return article_soup.find(name='div',attrs={'class':'content-inner-wrap'})

    def get_date(self,main_content):
        return main_content.find(name='span',attrs={'class':'article-date'}).text.split()[-1]

    def get_title(self,main_content):
        return main_content.find(name='h1',attrs={'class':'wi-article-title article-title-main'}).text.replace('\r','').replace('\n','').strip()

    def get_authors(self,main_content):
        authors_divs = main_content.find_all(name='a',
                                             attrs={'class': 'linked-name js-linked-name stats-author-info-trigger'})
        authors = ', '.join([elt.text for elt in authors_divs])
        return authors

    def get_doi(self,main_content):
        return main_content.find(name='div', attrs={'class': 'citation-doi'}).text.replace('\n', '').strip()

    def get_abstract(self,main_content):
        abstract= main_content.find(name='section',attrs={'class':'abstract'})
        if abstract:
            return abstract.text
        return ''

    def get_keywords(self,main_content):
        keywords_divs = main_content.find_all(name='a', attrs={'class': 'kwd-part kwd-main'})
        keywords = ', '.join([elt.text for elt in keywords_divs])
        return keywords

    def get_numCitedBy(self,article_soup):
        numCitedBy_div = article_soup.find(name='div', attrs={'class': "article-cited-link-wrap web-of-science"})
        numCitedBy = .5
        if numCitedBy_div:
            numCitedBy_text = numCitedBy_div.text
            numCitedBy = re.findall('\d+', numCitedBy_text)[0]
        return numCitedBy

    def get_volumes_dirs_in_years(self,year1,year2):
        volumes_dirs = [self.get_year_volumes(yr) for yr in tqdm(range(year1,year2+1))]
        print(f'Total volumes : {len(volumes_dirs)}')
        return volumes_dirs

    def get_docs_from_volume(self,volume_dir):
        articles_soups = self.get_articles_soups_in_volume(volume_dir=volume_dir)
        for idx,soup in enumerate(articles_soups):
            main_content = self.get_main_content_article(soup)
            if main_content:
                abstract = self.get_abstract(main_content)
                if abstract:
                    title = self.get_title(main_content)
                    authors = self.get_authors(main_content)
                    date = self.get_date(main_content)
                    doi = self.get_doi(main_content)
                    keywords = self.get_keywords(main_content)
                    numCitedBy = self.get_numCitedBy(soup)

                    self.docs['title'].append(title)
                    self.docs['abstract'].append(abstract)
                    self.docs['authors'].append(authors)
                    self.docs['date'].append(date)
                    self.docs['field'].append('Environmental Science')
                    self.docs['keywords'].append(keywords)
                    self.docs['doi'].append(doi)
                    self.docs['numCitedBy'].append(numCitedBy)
            else:
                print('no main content in idx : ', idx)


    def get_docs(self,year1,year2):

        #Get articles pages
        print('>> Getting volumes dirs..')
        volumes_dirs=self.get_volumes_dirs_in_years(year1=year1,year2=year2)
        volumes_dirs = [elt for elt2 in volumes_dirs for elt in elt2]

        print('>> Getting papers..')
        for volume_dir in tqdm(volumes_dirs):
            self.get_docs_from_volume(volume_dir)





import os
import numpy as np
import faiss
from tqdm.notebook import tqdm
from naimai.utils.general import load_gzip, save_gzip
from naimai.constants.fields import all_fields
from naimai.constants.paths import path_dispatched, path_similarity_model
from sentence_transformers import SentenceTransformer

class Dispatcher:
    '''
    Takes the formatted papers of a database and dispatches papers over the fields.
    Principle : Similarity between the papers data (fields or keywords) with the fields encoded & stored using faiss.
    '''
    def __init__(self, papers_dict, top_n=3, model=None, fields_index=None):
        self.papers = papers_dict
        self.top_n = top_n
        self.model = model
        self.fields_index = fields_index
        self.fields_elements = {k: {} for k in all_fields}

    def load_model(self, path=path_similarity_model):
        if not self.model:
            self.model = SentenceTransformer(path)

    def get_fields_index(self):
        encoded_fields = self.model.encode(all_fields)
        encoded_fields = np.asarray(encoded_fields.astype('float32'))
        self.fields_index = faiss.IndexIDMap(faiss.IndexFlatIP(768))
        self.fields_index.add_with_ids(encoded_fields, np.array(range(len(all_fields))))

    def query_from_fields(self, fields):
        return ', '.join(fields)

    def query_from_keywords(self, paper):
        return paper['Keywords']

    def query_from_title(self, paper):
        return paper['Title']

    def get_query(self, paper):
        fields = [elt for elt in paper['fields'] if isinstance(elt, str)]
        if len(fields) > 2:
            query = self.query_from_fields(fields)
        else:
            query = self.query_from_fields(fields) + ' ' + self.query_from_keywords(paper) + ' ' + self.query_from_title(
                paper)
        return query

    def paper2fields(self, paper):
        query = self.get_query(paper)
        query_vector = self.model.encode([query])
        ids = self.fields_index.search(query_vector, self.top_n)[1].tolist()[0]
        fields = list(np.array(all_fields)[ids])
        # paths = [os.path.join(path_dispatched,field) for field in fields]
        return fields

    def update_fields_elements(self, fname, paper, paper_fields):
        for field in paper_fields:
            self.fields_elements[field].update({fname: paper})

    def save_field_elements(self, field, file_name='all_papers', update=False):
        papers_field = self.fields_elements[field]
        path_field = os.path.join(path_dispatched, field, file_name)
        if update and os.path.exists(path_field):
            loaded_papers = load_gzip(path_field)
            loaded_papers.update(papers_field)
            save_gzip(path_field, loaded_papers)
        else:
            save_gzip(path_field, papers_field)

    def save_elements(self,file_name='all_papers', update=False):
        for field in tqdm(self.fields_elements):
            if self.fields_elements[field]:
                self.save_field_elements(field=field, file_name=file_name,update=update)

    def dispatch(self, save=False,file_name='all_papers', update=False):
        print('>> Initialization..')
        self.load_model()
        self.get_fields_index()
        print('>> Dispatching..')
        for fname in tqdm(self.papers):
            paper = self.papers[fname]
            paper_fields = self.paper2fields(paper)
            self.update_fields_elements(fname, paper, paper_fields)
        if save:
            print('>> Saving..')
            self.save_elements(update=update,file_name=file_name)
            print('>> Finished dispatching!')







import os
from google.cloud import storage
from naimai.constants.paths import path_produced, path_bomr_classifier, path_objective_classifier


class gcloud_data:
    def __init__(self, bucket_name="naima_bucket2", path_service_key_gcloud='service_key_gcloud.json',
                 prefix='data_fields/'):
        self.bucket_name = bucket_name
        self.storage_client = None
        self.bucket = None
        self.prefix = prefix

        os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = path_service_key_gcloud
        try:
            self.get_client()
            self.get_bucket()
        except:
            print('No storage client found! Give the correct path of service key gcloud file.')

    def get_client(self):
        self.storage_client = storage.Client()

    def get_bucket(self):
        if self.storage_client:
            self.bucket = self.storage_client.get_bucket(self.bucket_name)
        else:
            print('Need storage client!')

    def upload(self, path_from, path_to):
        blob = self.bucket.blob(path_to)
        blob.upload_from_filename(path_from)

    def upload_search_model(self, field: str, path_drive: str):
        '''
        upload search model folder from drive to bucket
        '''
        path_drive2 = os.path.join(path_drive, 'search_model')
        path_gcloud = os.path.join(self.prefix, field, 'search_model')

        files = os.listdir(path_drive2)
        path_drive_pooling = ''
        # upload files
        for f in files:
            path = os.path.join(path_drive2, f)
            if os.path.isfile(path):
                path_file_gcloud = os.path.join(path_gcloud,f)
                self.upload(path, path_file_gcloud)
            elif os.path.isdir(path):
                path_drive_pooling = os.path.join(path, 'config.json')

        # upload pooling folder
        path_gcloud_pooling = os.path.join(path_gcloud, '1_Pooling', 'config.json')
        self.upload(path_drive_pooling, path_gcloud_pooling)

    def upload_folder(self,path_folder: str,dir_gcloud=''):
        '''
        upload folder containing files from drive to bucket
        :param path_folder:
        :return:
        '''
        files_ = os.listdir(path_folder)
        path_files = [os.path.join(path_folder,fil) for fil in files_]
        for path,f in zip(path_files,files_):
            path_gcloud = os.path.join(dir_gcloud,f)
            self.upload(path,path_gcloud)

    def upload_classifiers(self):
        '''
        upload obj classifier and bomr classifier for custom queries
        :return:
        '''
        obj_clf_folder = path_objective_classifier.split('/')[-1]
        bomr_clf_folder = path_bomr_classifier.split('/')[-1]
        dir_gcloud = 'models/'

        dir_gcloud_obj = os.path.join(dir_gcloud,obj_clf_folder)
        dir_gcloud_bomr = os.path.join(dir_gcloud, bomr_clf_folder)

        print('>> Uploading obj classifier..')
        self.upload_folder(path_objective_classifier,dir_gcloud_obj)
        print('>> Uploading bomr classifier..')
        self.upload_folder(path_bomr_classifier,dir_gcloud_bomr)
        print('>> Done!')


    def upload_papers(self, field: str, path_drive: str):
        '''
        upload papers sqlite data from drive to bucket
        '''
        path_drive2 = os.path.join(path_drive, 'all_papers_sqlite')
        path_gcloud = os.path.join(self.prefix, field, 'all_papers_sqlite')
        self.upload(path_drive2, path_gcloud)

    def upload_index(self, field: str, path_drive: str):
        '''
        upload field index from drive to bucket
        '''
        path_drive2 = os.path.join(path_drive, 'encodings.index')
        path_gcloud = os.path.join(self.prefix, field, 'encodings.index')
        self.upload(path_drive2, path_gcloud)

    def upload_field(self, field: str,upload_papers=True,upload_field_index=True,upload_search_model=True):
        '''
        upload field from drive to bucket
        '''
        path_drive_field = os.path.join(path_produced, field)
        if upload_papers:
            print('>> Uploading papers..')
            self.upload_papers(field, path_drive_field)

        if upload_field_index:
            print('>> Uploading index..')
            self.upload_index(field, path_drive_field)

        if upload_search_model:
            print('>> Uploading search model..')
            self.upload_search_model(field, path_drive_field)
        print('>> Done !')

    def get_folder(self,dir_gcloud,dir_local):
        '''
        get folder from gcloud
        :return:
        '''
        blobs = self.bucket.list_blobs(prefix=dir_gcloud)

        if not os.path.exists(dir_local):
            os.makedirs(dir_local,exist_ok=True)
            for blob in blobs:
                blob.download_to_filename(blob.name)

    def get_classifiers(self):
        '''
        get obj classifier and bomr classifier for custom queries
        :return:
        '''
        obj_clf_folder = 'distil_bert_obj_classifier'
        bomr_clf_folder = 'bigbird_roberta16'
        dir_gcloud = 'models/'

        dir_obj = os.path.join(dir_gcloud, obj_clf_folder)
        dir_bomr = os.path.join(dir_gcloud, bomr_clf_folder)

        print('>> Getting obj classifier..')
        self.get_folder(dir_obj,dir_obj)
        print('>> Getting bomr classifier..')
        self.get_folder(dir_bomr,dir_bomr)
        print('>> Done!')

    def get_index(self, field: str):
        '''
        get index from gcloud
        :param field:
        :return:
        '''
        path_index = os.path.join(self.prefix, field, 'encodings.index') # same path in gcloud and locally
        blobs = self.bucket.list_blobs(prefix=path_index)

        if not os.path.exists(path_index): #if not already downloaded locally..
            path = os.path.join(self.prefix, field)
            os.makedirs(path,exist_ok=True)
            for blob in blobs:
                blob.download_to_filename(path_index)

    def get_papers(self, field: str):
        '''
        get papers sqlite from gcloud
        :param field:
        :return:
        '''
        path_papers = os.path.join(self.prefix, field, 'all_papers_sqlite') # same path in gcloud and locally
        blobs = self.bucket.list_blobs(prefix=path_papers)

        if not os.path.exists(path_papers): #if not already downloaded locally..
            path = os.path.join(self.prefix, field)
            os.makedirs(path,exist_ok=True)
            for blob in blobs:
                blob.download_to_filename(path_papers)

    def get_search_model(self, field: str):
        '''
        get search model from gcloud
        :param field:
        :return:
        '''
        path_smodel = os.path.join(self.prefix, field, 'search_model') # same path in gcloud and locally
        blobs = self.bucket.list_blobs(prefix=path_smodel)

        if not os.path.exists(path_smodel): #if not already downloaded locally..
            os.makedirs(path_smodel,exist_ok=True)
            path_pooling = os.path.join(path_smodel,'1_Pooling')
            os.mkdir(path_pooling)
            for blob in blobs:
                blob.download_to_filename(blob.name)

    def get_files(self,field: str, upload_papers=True,upload_field_index=True,upload_search_model=True):
        '''
        get papers, index and search model
        :param field:
        :return:
        '''
        if upload_papers:
            print('>> Getting papers..')
            self.get_papers(field)

        if upload_field_index:
            print('>> Getting index..')
            self.get_index(field)

        if upload_search_model:
            print('>> Getting search model..')
            self.get_search_model(field)
        print('>> Done downloading!')


'''
The EGU papers could not be processed with papers class (in papers/raw.py) because of the doi, fields and journal data.
So the get_doi, get_fields and get_journal methods are overwritten in paper_egu class, and the add_paper method is
overwritten in papers_egu to use paper_egu (instead of paper_base).
'''

from naimai.papers.raw import papers, paper_base

class paper_egu(paper_base):
    '''
    Paper class that map a row if the csv file with data about papers to a dictionary.
    '''
    def __init__(self,df,idx_in_df: int):
        super().__init__(df,idx_in_df)

    def get_doi(self):
        if str(self.paper_infos['doi'])=='nan':
            self.doi=''
        else:
            self.doi = self.paper_infos['doi'].replace('https://doi.org/','')

    def get_fields(self):
        self.fields = [self.paper_infos['field'],]

    def get_journal(self):
        self.Journal =  'Atmospheric Chemistry and Physics'


class papers_egu(papers):
    def __init__(self,papers_path: str,database,nlp=None):        
        '''
        inherits from the papers class
        :param papers_path: path of the csv file
        :param database: name of the database if all the papers in the csv file are coming from the same source.
        :param nlp:  spaCy nlp pipeline
        '''
        super().__init__(papers_path,database,nlp) # loading self.naimai_dois & other attributes

    def add_paper(self,idx_in_data: int,check_database=True):
        '''
        Add a paper data as element
        :param idx_in_data: idx of the paper in the csv file
        :param check_database: if True, the paper is not added as element if already contained in the database.
        :return:
        '''
        new_paper = paper_egu(df=self.data,
                                idx_in_df=idx_in_data)
        new_paper.get_doi()
        if new_paper.doi:
            new_paper.get_Title()
            if check_database:
                is_in_database_condition = new_paper.is_in_database(self.naimai_dois)
            else:
                is_in_database_condition = False
            if not is_in_database_condition:
                new_paper.get_Abstract()
                if len(new_paper.Abstract.split())>5:
                    new_paper.database = self.database
                    new_paper.get_fields()
                    new_paper.get_journal()
                    new_paper.get_Authors()
                    new_paper.get_year()
                    new_paper.replace_abbreviations()
                    new_paper.get_numCitedBy()
                    self.elements[new_paper.doi] = new_paper.save_dict()
                    self.naimai_dois.append(new_paper.doi)




import re
from tqdm.notebook import tqdm
from bs4 import BeautifulSoup
import random
import requests
import time

class IAHR_Crawler:
    def __init__(self, path,t_min=1,t_max=3):
        self.path = path
        self.papers_paths = []
        self.docs = {'title': [], 'abstract': [], 'authors': [], 'date': [], 'field': [], "keywords":[],"webpage":[]}
        self.t_min=t_min
        self.t_max=t_max

    def get_soup(self, path):
        header = {}
        header[
            'User-Agent'] = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'
        soup = BeautifulSoup(requests.get(path, headers=header, timeout=15).content, 'html.parser')
        slp=random.randint(self.t_min,self.t_max)
        time.sleep(slp)
        return soup


    def get_abstract(self,body):
        abstract = body[3].text.replace('Abstract: ', '').strip()
        return abstract

    def get_paths_in_page(self, nb_page):
        webpage = self.path +'&page='+str(nb_page)
        soup = self.get_soup(webpage)
        paper_list = soup.find_all(name='tbody')[0]
        papers = paper_list.find_all(name='tr')
        papers_pages = [elt.find(name='a', attrs={'class': 'overflow-text2'})['href'] for elt in papers]
        paths_in_page = ['https://www.iahr.org' + elt for elt in papers_pages]
        return paths_in_page

    def get_papers_pages(self,first_page,last_page):
        for page in tqdm(range(first_page,last_page+1)):
            paths_in_page = self.get_paths_in_page(page)
            self.papers_paths +=paths_in_page

    def get_paper_path_content(self,paper_path):
        soup = self.get_soup(paper_path)
        content = soup.find_all(name='div', attrs={'class': 'content-text'})[1]
        return content

    def get_content_body(self,content):
        body = content.find(name='div', attrs={'class': None}).find_all(name='p')
        return body

    def get_keywords(self,body):
        keywords = body[2].text.replace('Keywords: ', '').replace(';',',').strip()
        return keywords

    def get_title(self, content):
        title = content.find(name='h2').text
        return title

    def get_abstract_id(self, description_soup):
        return description_soup.find_all(name="a", attrs={"class": "title optClickTitle"})[0]['href'].split('=')[1]

    def get_date(self, body):
        date = body[5].text.replace('Year: ', '').strip()
        return date

    def get_authors(self, body):
        authors = body[0].text.replace('\nAuthor(s): ','').strip()
        return authors

    def get_docs(self,first_page,last_page):

        #Get articles pages
        print('>> Getting pages..')
        self.get_papers_pages(first_page=first_page,last_page=last_page)
        # self.docs = {'title': [], 'abstract': [], 'authors': [], 'date': [], 'field': [], "keywords": [], "webpage": []}

        print('>> Getting papers..')
        #crawl papers
        for path in tqdm(self.papers_paths):
            content = self.get_paper_path_content(path)
            title = self.get_title(content)
            body = self.get_content_body(content)
            abstract = self.get_abstract(body)
            if abstract:
                authors = self.get_authors(body)
                keywords = self.get_keywords(body)
                date = self.get_date(body)

                self.docs['title'].append(title)
                self.docs['abstract'].append(abstract)
                self.docs['authors'].append(authors)
                self.docs['date'].append(date)
                self.docs['field'].append('Environmental Science')
                self.docs['keywords'].append(keywords)
                self.docs['webpage'].append(path)
